# coding: utf-8

import datetime
import logging

from nova.api import validation
from nova.api.openstack import extensions, wsgi
from webob.exc import HTTPBadRequest

from tionix_client.api.helpers import (
    get_instances_by_project, tionix_admin_guard)
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.journal.manager import JournalManager
from tionix_client.journal.constants import ObjectType
from tionix_client.api.compute.schemas.journal import create_journal

logger = logging.getLogger(__name__)

ALIAS = 'tnx-journal'

authorize = extensions.os_compute_authorizer(ALIAS)


def get_bool_from_request(req, param, default_value=None):

    value = req.GET.get(param, default_value)
    # noinspection PyTypeChecker
    if isinstance(value, basestring):
        value = value.lower()
    value = value in (True, 'true')

    return value


class TnxJournalController(wsgi.Controller):

    _default_limit = 10
    _default_offset = 0

    _journal_attrs = (
        'id', 'object_id', 'object_type', 'action', 'status',
        'is_notification', 'event_time', 'action_arguments', 'response_code',
        'message', 'user_id', 'user_name', 'request_id', 'parent_request_id')

    def _as_dict(self, journl_row):
        journal_dict = {}

        for journal_attr in self._journal_attrs:
            attr = getattr(journl_row, journal_attr, '')
            if isinstance(attr, datetime.datetime):
                attr = attr.isoformat()
            journal_dict[journal_attr] = attr
        return journal_dict

    @extensions.expected_errors((402,))
    def index(self, req):
        """Return journal logs by filter."""

        project_id = req.GET.pop('project_id', None)
        user_id = req.GET.pop('user_id', None)
        object_type = req.GET.pop('object_type', None)
        object_ids = req.GET.getall('object_ids')
        offset = req.GET.pop('offset', None)
        limit = req.GET.pop('limit', None)
        show_all = get_bool_from_request(req, 'all', default_value=False)

        if isinstance(offset, (str, unicode)):
            offset = int(offset) if offset.isdigit() else self._default_offset

        if isinstance(limit, (str, unicode)):
            limit = int(limit) if limit.isdigit() else self._default_limit

        kwargs = {
            'limit': limit,
            'offset': offset,
        }
        if user_id:
            kwargs['user_id'] = user_id
        if object_type:
            kwargs['object_type'] = object_type
        if object_ids:
            kwargs['object_id__in'] = object_ids
        elif project_id and object_type == ObjectType.Server:
            kwargs['object_id__in'] = get_instances_by_project(req, project_id)
        if not show_all:
            kwargs['parent_request_id'] = None

        journal = JournalManager.get_many(**kwargs)
        result = [self._as_dict(journal_row) for journal_row in journal]
        return {'journal': result}

    # noinspection PyUnusedLocal
    @wsgi.response(201)
    @extensions.expected_errors((400, 402, 403, 409))
    @validation.schema(create_journal)
    @tionix_admin_guard
    def create(self, req, body):
        """Create a new journal record."""
        journal = body['journal']

        try:
            journal_row = JournalManager.log_action(
                is_notification=False,
                event_time=datetime.datetime.utcnow(),
                **journal)
        except Exception as err:
            msg = _(u"Couldn't create journal record: {0}").format(
                unicode(err))
            raise HTTPBadRequest(explanation=msg)
        return {'journal': self._as_dict(journal_row)}


class TnxJournal(extensions.V21APIExtensionBase):
    """TIONIX Journal extension"""
    name = 'TnxJournalExtension'
    alias = ALIAS
    version = 1

    def get_resources(self):
        resources = [
            extensions.ResourceExtension(
                self.alias,
                TnxJournalController())
        ]
        return resources

    def get_controller_extensions(self):
        return []
