# coding: utf-8
# pylint: disable=too-many-lines

import argparse
import json
import logging
import logging.config
import os
import tempfile
import time
import warnings

from collections import defaultdict

import pymysql
from pymysql import cursors
from pymysql.err import OperationalError

from neutronclient.common.exceptions import NotFound
from novaclient.exceptions import NotFound as NovaNotFound

from tionix_client.os_clients.keystone_client import get_keystone_client
from tionix_client.os_clients.nova_client import get_nova_client
from tionix_client.os_clients.neutron_client import get_neutron_client
from tionix_client.os_clients.cinder_client import get_cinder_client
from tionix_client.settings import CONF


# Reload CONF to not overide log level when reading
CONF.reload(update_loglevel=False)


logger = logging.getLogger()


SERVER_CHECK_SLEEP = 6
SERVER_CHECK_TRIES = 10


# noinspection SqlNoDataSourceInspection
class DBManager(object):
    """SQL database manager"""

    table_name = None
    sql_fields = None

    SQL_CREATE_TABLE = (
        'CREATE TABLE IF NOT EXISTS `{table_name}` ('
        '`id` INT(11) NOT NULL AUTO_INCREMENT, '
        '{sql_fields}, '
        'PRIMARY KEY(`id`))')

    SQL_INSERT = 'INSERT INTO `{table_name}` ({fields}) VALUES ({values})'
    SQL_SELECT_WHERE = (
        'SELECT * FROM `{table_name}` WHERE {search_params} LIMIT {limit}')
    SQL_SELECT = 'SELECT * FROM `{table_name}`'
    SQL_UPDATE = 'UPDATE `{table_name}` SET {field_values} WHERE `id`=%s'

    def __init__(self, host, port, db_name, username, password):
        self.json_fields = []
        for key, value in self.sql_fields.items():
            if value == 'JSON':
                # pylint: disable=unsupported-assignment-operation
                self.sql_fields[key] = 'TEXT'
                self.json_fields.append(key)

        self.db_params = dict(
            host=host,
            port=port,
            user=username,
            password=password,
            db=db_name,
            charset='utf8mb4',
            cursorclass=cursors.DictCursor
        )
        self.connection = pymysql.connect(**self.db_params)
        self._ensure_table()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()

    def _ensure_table(self):
        """Create table if not exists"""
        fields = ', '.join(
            ['`{0}` {1}'.format(k, v) for k, v in self.sql_fields.items()])

        sql = self.SQL_CREATE_TABLE.format(
            table_name=self.table_name,
            sql_fields=fields)

        with self.connection.cursor() as cursor:
            cursor.execute(sql)
        self.connection.commit()

    def _deserialize(self, instance):
        body = dict(**instance)
        if body:
            for k, v in body.items():
                if k in self.json_fields and v:
                    body[k] = json.loads(v)
        return body

    def _serialize(self, instance):
        # pylint: disable=not-an-iterable
        body = dict(**instance)
        if body:
            for k, v in body.items():
                if k in self.json_fields and v is not None:
                    body[k] = json.dumps(v)
        for field in self.sql_fields:
            if field not in body:
                body[field] = None
        return body

    def create(self, **params):
        # pylint: disable=not-an-iterable
        instance = self._serialize(params)
        fields = ', '.join(['`{0}`'.format(k) for k in self.sql_fields])
        values = ', '.join(['%s'] * len(instance.values()))

        sql = self.SQL_INSERT.format(
            table_name=self.table_name,
            fields=fields,
            values=values)

        with self.connection.cursor() as cursor:
            cursor.execute(sql, instance.values())

        self.connection.commit()
        return self.get_one(id=cursor.lastrowid)

    def get_many(self, limit=10, **params):
        query_params = params.values()
        if params:
            search_params = ' AND '.join(
                ['{0}=%s'.format(key) for key in params])
            sql = self.SQL_SELECT_WHERE.format(
                table_name=self.table_name, search_params=search_params,
                limit=limit)
        else:
            sql = self.SQL_SELECT.format(
                table_name=self.table_name)

        with self.connection.cursor() as cursor:
            cursor.execute(sql, query_params)

        return [self._deserialize(m) for m in cursor.fetchall()]

    def get_one(self, **params):
        results = self.get_many(limit=1, **params)
        return results[0] if results else None

    # noinspection PyTypeChecker
    def update(self, instance):
        body = self._serialize(instance)
        instance_id = body.pop('id')

        # pylint: disable=not-an-iterable
        field_values = ', '.join(
            ['{0}=%s'.format(k) for k in self.sql_fields])

        sql = self.SQL_UPDATE.format(
            table_name=self.table_name,
            field_values=field_values)

        with self.connection.cursor() as cursor:
            cursor.execute(sql, body.values() + [instance_id])
        self.connection.commit()


class MigrationsManager(DBManager):
    table_name = 'tnx_migrations'

    sql_fields = {
        'server_id': 'VARCHAR(36) NOT NULL UNIQUE',
        'status': 'VARCHAR(36)',
        'migration_steps': 'JSON',
        'context': 'JSON',
    }


class MigrationStatus(object):
    PROGRESS = 'progress'
    COMPLETED = 'completed'
    ERROR = 'error'
    REVERTED = 'reverted'


class CommandError(Exception):
    pass


class BaseRunner(object):
    def __init__(self, db_manager, args):
        self.db_manager = db_manager
        self.args = args

        self.cache = {'servers': {}, 'volumes': {}}

        opts = {
            'username': args.os_username,
            'password': args.os_password,
            'tenant': args.os_project_name,
            'user_domain_name': args.os_user_domain_name,
            'project_domain_name': args.os_project_domain_name,
            'auth_url': args.os_auth_url,
            'auth_version': args.os_auth_version,
        }

        self.keystone_client = ks_client = get_keystone_client(**opts)
        self.nova_client = get_nova_client(keystone_client=ks_client)
        self.neutron_client = get_neutron_client(keystone_client=ks_client)
        self.cinder_client = get_cinder_client(keystone_client=ks_client)

    def run(self):
        raise NotImplementedError

    def get_server(self, server_id, cached=True):
        server = self.cache['servers'].get(server_id) if cached else None
        if not server:
            server = self.nova_client.servers.get(server_id)
            self.cache['servers'][server_id] = server
        return server

    def get_volume(self, volume_id, cached=True):
        volume = self.cache['volumes'].get(volume_id) if cached else None
        if not volume:
            volume = self.cinder_client.volumes.get(volume_id)
            self.cache['volumes'][volume_id] = volume
        return volume

    def get_server_ports_info(self, server_id):
        port_keys = (
            'admin_state_up',
            'allowed_address_pairs',
            'binding:host_id',
            'binding:profile',
            'binding:vnic_type',
            'description',
            'extra_dhcp_opts',
            'fixed_ips',
            'mac_address',
            'network_id',
            'port_security_enabled',
            'security_groups',
            'tenant_id'
        )

        server = self.get_server(server_id)
        ports = defaultdict(dict)
        for interface in server.interface_list():
            port = self.neutron_client.show_port(interface.port_id)['port']
            ports[port['id']].update({key: port.get(key) for key in port_keys})
        return ports

    def get_server_volumes(self, server_id):
        server = self.get_server(server_id)
        volumes = {}
        root_device_name = getattr(server, 'OS-EXT-SRV-ATTR:root_device_name')

        for volume_info in getattr(server,
                                   'os-extended-volumes:volumes_attached', []):
            volume = self.get_volume(volume_info['id'])
            for attachment in volume.attachments:
                if attachment['server_id'] == server.id:
                    volumes[volume.id] = {
                        'id': volume.id,
                        'name': volume.name,
                        'size': volume.size,
                        'attachment_id': attachment['id'],
                        'device': attachment['device'],
                        'is_root': root_device_name == attachment['device'],
                        'volume_type': volume.volume_type
                    }
                    break
            else:
                continue

        return volumes

    def wait_server_deletion(self, server_id):
        i = 0
        while i < SERVER_CHECK_TRIES:
            logger.info(
                u'Waiting for server %s to be deleted.',
                server_id)
            try:
                server = self.get_server(server_id, cached=False)
            except NovaNotFound:
                break
            logger.info(u'Server status: {0}'.format(server.status))
            time.sleep(SERVER_CHECK_SLEEP)
            i += 1
        else:
            raise CommandError(
                u'Failed to delete server {0}.'.format(server_id))


class MigrationRunner(BaseRunner):
    @staticmethod
    def log_result(context):
        attach_volumes = context.get('attach_volumes')
        server_id = context['new_server_id'] or context['server_id']
        root_volume = None
        for volume in attach_volumes:
            if volume['is_root']:
                root_volume = volume
                break
        attach_volumes.remove(root_volume)
        msg = u"Result:\ninstance_uuid:{0}\nroot_volume:{1} {2}\n".format(
            server_id, root_volume['id'], root_volume['volume_type'])
        msg += u'Attached volumes:\n'
        for v in attach_volumes:
            msg += u"{0} {1}\n".format(v['id'], v['volume_type'] or '-')
        logger.info(msg)

    @staticmethod
    def check_step(migration, step_name, check):
        result = False
        for step in migration['migration_steps']:
            (step, content), = step.items()
            if step == step_name and check(content):
                result = True
                break
        return result

    def make_context(self, server):
        replace_volumes = {}
        for volume_param in self.args.volume:
            volume_id, volume_type = volume_param.split(':')
            self.get_volume(volume_id)
            replace_volumes[volume_id] = volume_type

        server_volumes = self.get_server_volumes(server.id)
        root_volume = next(
            (v['id'] for v in server_volumes.values()
             if v['is_root']), None)

        if not root_volume and not self.args.root_volume_type:
            raise CommandError(
                u'--root-volume-type argument is required '
                u'for ephemeral instances.')

        ports_info = self.get_server_ports_info(server.id)
        if not ports_info:
            raise CommandError(
                u'Unable to migrate server without network interfaces.')

        context = {
            'server_id': server.id,
            'new_server_id': None,
            'server_volumes': server_volumes,
            'root_volume_id': root_volume,
            'root_volume_type': self.args.root_volume_type,
            'flavor_id': server.flavor['id'],
            'new_volumes': [],
            'replace_volumes': replace_volumes,
            'ports_info': ports_info,
            'networks': [],
            'attach_volumes': [],
        }
        return context

    def run(self):
        server = self.get_server(self.args.server)
        if self.keystone_client.auth_tenant_id != server.tenant_id:
            raise CommandError(
                u"Server tenant ID doesn't match authenticated tenant ID.")

        if not server.status.lower() == 'shutoff':
            raise CommandError(
                u'Instance is expected to be in SHUTOFF status.')

        migration = self.db_manager.get_one(server_id=server.id)

        if self.args.retry:
            if not migration:
                raise CommandError(u'No migration found for retry.')

            if migration['status'] == MigrationStatus.REVERTED:
                raise CommandError(u'Unable to retry reverted migration.')

            context = migration.get('context')
            if not context:
                raise CommandError(
                    u'Unable to retry: migration context is empty.')
        else:
            if migration and migration['status'] not in (
                    MigrationStatus.COMPLETED, MigrationStatus.REVERTED):
                raise CommandError(
                    u'There is incomplete migration with status {status}. '
                    u'Run command with `--retry` '
                    u'or `--force-complete`.'.format(
                        status=migration['status']))
            context = self.make_context(server)

        steps = [
            'volume.create',
            'volume.rename',
            'volume.detach',
        ]

        root_volume = context.get('root_volume_id')
        replace_volumes = context.get('replace_volumes') or []

        if root_volume and root_volume not in replace_volumes:
            steps += ['volume.attach']
        else:
            steps += [
                'port.detach',
                'port.create',
                'server.create',
                'server.rename',
            ]

        if migration:
            migration['status'] = MigrationStatus.PROGRESS
            if not self.args.retry:
                migration['migration_steps'] = []
            migration['context'] = context
            self.db_manager.update(migration)
        else:
            migration = self.db_manager.create(
                server_id=server.id,
                status=MigrationStatus.PROGRESS,
                context=context,
                migration_steps=[])

        logger.info(u'Migrating server %s', self.args.server)
        try:
            for i, step in enumerate(steps, start=1):
                logger.info(u'Processing step %s/%s: %s', i, len(steps), step)
                func_name = 'step_' + step.replace('.', '_')
                migration_method = getattr(self, func_name, None)
                migration_method(context, migration)
        except Exception as err:
            if self.args.debug:
                logger.exception(err)
            migration['status'] = MigrationStatus.ERROR
            self.db_manager.update(migration)
            raise CommandError(u'Migration failed: {0}'.format(err))

        migration['status'] = MigrationStatus.COMPLETED
        self.db_manager.update(migration)
        self.log_result(context)

    def step_volume_create(self, context, migration):
        server_volumes = context.get('server_volumes')
        replace_volumes = context.get('replace_volumes')
        root_volume_id = context.get('root_volume_id')
        root_volume_type = context.get('root_volume_type')
        new_volumes = context.get('new_volumes')
        flavor_id = context.get('flavor_id')
        server_id = context.get('server_id')

        for original_id, volume_type in replace_volumes.items():

            if self.check_step(
                    migration, 'volume.create',
                    check=lambda c, val=original_id: c['source_id'] == val):
                logger.info(
                    u'Skip processing volume %s:%s', original_id, volume_type)
                continue

            volume = server_volumes.get(original_id)
            new_volume = self.cinder_client.volumes.create(
                name=volume['name'],
                size=volume['size'],
                volume_type=replace_volumes.get(original_id))
            if volume['is_root']:
                self.cinder_client.volumes.set_bootable(new_volume, True)

            volume_info = {
                'id': new_volume.id,
                'is_root': volume['is_root'],
                'source_id': volume['id'],
                'device': volume['device'],
                'volume_type': replace_volumes.get(original_id),
            }
            new_volumes.append(volume_info)

            migration['migration_steps'].append({'volume.create': volume_info})
            migration['context'] = context
            self.db_manager.update(migration)
            logger.info(
                u'Created volume: uuid=%s, name=%s, size=%s, type=%s',
                new_volume.id, volume['name'], volume['size'],
                replace_volumes.get(original_id))

        if not root_volume_id:
            if self.check_step(
                    migration, 'volume.create',
                    check=lambda c: c['is_root']):
                logger.info(u'Skip creating volume for the root disk.')
            else:
                flavor = self.nova_client.flavors.get(flavor_id)
                server = self.get_server(server_id)
                new_volume = self.cinder_client.volumes.create(
                    name=server.name + '_volume',
                    size=flavor.disk,
                    volume_type=root_volume_type)
                self.cinder_client.volumes.set_bootable(new_volume, True)

                volume_info = {
                    'id': new_volume.id,
                    'is_root': True,
                    'source_id': None,
                    'device': None,
                    'volume_type': root_volume_type,
                }
                new_volumes.append(volume_info)

                migration['migration_steps'].append(
                    {'volume.create': volume_info})
                migration['context'] = context
                self.db_manager.update(migration)
                logger.info(
                    u'Created volume for root disk: '
                    u'uuid=%s, name=%s, size=%s, type=%s',
                    new_volume.id, new_volume.name, flavor.disk,
                    root_volume_type)

    # noinspection PyArgumentList
    def step_volume_rename(self, context, migration):
        replace_volumes = context.get('replace_volumes')

        for volume_id in replace_volumes:

            if self.check_step(
                    migration, 'volume.rename',
                    check=lambda c, val=volume_id: c['id'] == val):
                logger.info(u'Skip renaming volume %s', volume_id)
                continue

            volume = self.get_volume(volume_id)
            volume.update(name=volume.name + '-old')

            migration['migration_steps'].append({
                'volume.rename': {
                    'id': volume.id,
                    'old_name': volume.name,
                    'new_name': volume.name + '-old'
                }
            })
            migration['context'] = context
            self.db_manager.update(migration)
            logger.info(
                u'Renamed volume with uuid=%s from %s to %s.',
                volume.id, volume.name, volume.name + '-old')

    def step_volume_detach(self, context, migration):
        server_volumes = context.get('server_volumes')
        replace_volumes = context.get('replace_volumes')
        new_volumes = context.get('new_volumes')
        root_volume_id = context.get('root_volume_id')
        server_id = context.get('server_id')

        if root_volume_id and root_volume_id not in replace_volumes:
            # Instance root disk is in the persistent storage and
            # no need to change type of the volume.
            # New instance will not be created.

            # Detach old volumes provided in --volumes argument.
            volumes_list = [
                v for v in server_volumes.values()
                if v['id'] in replace_volumes]

            context['attach_volumes'] = [
                v for v in server_volumes.values()
                if v['id'] not in replace_volumes] + new_volumes
        else:
            volumes_list = [
                v for v in server_volumes.values()
                if v['id'] not in replace_volumes]

            context['attach_volumes'] = volumes_list + new_volumes

        for volume_info in volumes_list:
            # pylint: disable=cell-var-from-loop
            volume_id = volume_info['id']
            if self.check_step(
                    migration, 'volume.detach',
                    check=lambda c: c['id'] == volume_id):
                logger.info(u'Skip detaching volume %s.', volume_id)
                continue

            self.nova_client.volumes.delete_server_volume(
                server_id, volume_info['attachment_id'])

            migration['migration_steps'].append({
                'volume.detach': {
                    'id': volume_id,
                    'from_server': server_id,
                    'device': volume_info['device'],
                }
            })
            migration['context'] = context
            self.db_manager.update(migration)
            logger.info(
                u'Detached volume with uuid=%s from instance with uuid=%s',
                volume_id, server_id)

    def step_volume_attach(self, context, migration):
        new_volumes = context.get('new_volumes')
        server_id = context.get('server_id')

        for volume_info in new_volumes:
            volume_id = volume_info['id']
            # pylint: disable=cell-var-from-loop
            if self.check_step(
                    migration, 'volume.attach',
                    check=lambda c: c['id'] == volume_id):
                logger.info(u'Skip attaching volume %s', volume_id)
                continue

            self.nova_client.volumes.create_server_volume(
                server_id, volume_id, device=volume_info['device'])

            migration['migration_steps'].append({
                'volume.attach': {
                    'id': volume_id,
                    'to_server': server_id,
                    'device': volume_info['device'],
                }
            })
            migration['context'] = context
            self.db_manager.update(migration)
            logger.info(
                u'Attached volume with uuid=%s to instance with uuid=%s. '
                u'Mount point: %s',
                volume_id, server_id, volume_info['device'])

    def step_port_detach(self, context, migration):
        ports_info = context.get('ports_info')
        server_id = context.get('server_id')

        server = self.get_server(server_id)

        for port_id in ports_info:

            port_info = ports_info[port_id]
            if self.check_step(
                    migration, 'port.detach',
                    check=lambda c, val=port_info: c['port_info'] == val):
                logger.info(u'Skip detaching port %s.', port_id)
                continue

            server.interface_detach(port_id)
            try:
                self.neutron_client.delete_port(port_id)
            except NotFound:
                pass

            migration['migration_steps'].append({
                'port.detach': {
                    'from_server': server.id,
                    'port_info': ports_info[port_id],
                }
            })
            migration['context'] = context
            self.db_manager.update(migration)

            logger.info(
                u'Detached port with uuid=%s from server with uuid=%s',
                port_id, server.id)

    def step_port_create(self, context, migration):
        networks = context.get('networks')
        ports_info = context.get('ports_info')

        for port_id, port_body in ports_info.items():
            port_info = ports_info[port_id]
            if self.check_step(
                    migration, 'port.create',
                    check=lambda c, val=port_info: c['port_info'] == val):
                logger.info(u'Skip creating port.')
                continue

            new_port = self.neutron_client.create_port(
                body={'port': port_body})

            migration['migration_steps'].append({
                'port.create': {
                    'id': new_port['port']['id'],
                    'port_info': port_body,
                }
            })
            networks.append({'port-id': new_port['port']['id']})
            migration['context'] = context
            self.db_manager.update(migration)

            logger.info(
                u'Created network port with uuid=%s', new_port['port']['id'])

    def step_server_create(self, context, migration):
        # pylint: disable=too-many-nested-blocks
        server_volumes = context.get('server_volumes')
        replace_volumes = context.get('replace_volumes')
        new_volumes = context.get('new_volumes')
        server_id = context.get('server_id')
        networks = context.get('networks')

        for step in migration['migration_steps']:
            (step, content), = step.items()
            if step == 'server.create':
                try:
                    server = self.get_server(content['id'], cached=False)
                except NovaNotFound:
                    pass
                else:
                    if server.status.lower == 'error':
                        logger.info(u'Server %s is in ERROR state.', server_id)
                        server.delete()
                        self.wait_server_deletion(server.id)
                    else:
                        logger.info(u'Skip creating server.')
                        return
                break

        server = self.get_server(server_id)
        volumes_list = [
            v for v in server_volumes.values()
            if v['id'] not in replace_volumes] + new_volumes

        block_device_mapping_v2 = []
        for volume_info in volumes_list:
            block_device = {
                "source_type": "volume",
                "uuid": volume_info['id'],
                "destination_type": "volume",
                "boot_index": 0 if volume_info['is_root'] else None
            }
            if volume_info['device']:
                block_device['device_name'] = volume_info['device']
            block_device_mapping_v2.append(block_device)

        new_server = self.nova_client.servers.create(
            name=server.name,
            image='',
            flavor=server.flavor['id'],
            nics=networks,
            block_device_mapping_v2=block_device_mapping_v2
        )

        migration['migration_steps'].append({
            'server.create': {
                'id': new_server.id
            }
        })
        context['new_server_id'] = new_server.id
        migration['context'] = context
        self.db_manager.update(migration)

        logger.info(
            u'Created new server with uuid=%s, name=%s',
            new_server.id, new_server.name)

    # noinspection PyArgumentList
    def step_server_rename(self, context, migration):
        server_id = context.get('server_id')

        if self.check_step(
                migration, 'server.rename',
                check=lambda c: c['id'] == server_id):
            logger.info(u'Skip renaming server %s', server_id)
            return

        server = self.get_server(server_id)
        server.update(name=server.name + '-old')

        migration['migration_steps'].append({
            'server.rename': {
                'id': server.id,
                'old_name': server.name,
                'new_name': server.name + '-old'
            }
        })
        migration['context'] = context
        self.db_manager.update(migration)
        logger.info(
            u'Renamed server with uuid=%s from %s to %s',
            server.id, server.name, server.name + '-old')


# noinspection PyArgumentList
class RevertRunner(BaseRunner):

    def run(self):
        migration = self.db_manager.get_one(server_id=self.args.server)
        if not migration:
            raise CommandError(
                u'Failed to find migration for server {0}'.format(
                    self.args.server))

        logger.info(u'Reverting migration for server %s', self.args.server)
        steps_count = len(migration['migration_steps'])
        for i, step in enumerate(migration['migration_steps'][::-1], start=1):
            (step_name, context), = step.items()
            func_name = 'revert_' + step_name.replace('.', '_')
            revert_method = getattr(self, func_name, None)

            if not revert_method:
                raise CommandError(
                    u'Failed to revert migration step: '
                    u'unknown step name "{0}"'.format(step_name))

            logger.info(u'Reverting step %s/%s: %s', i, steps_count, step_name)
            try:
                revert_method(context)
            except Exception as err:
                logger.error(u'Failed to revert step %s: %s', step_name, err,
                             exc_info=self.args.debug)

        migration['status'] = MigrationStatus.REVERTED
        self.db_manager.update(migration)

    def revert_volume_create(self, context):
        volume_id = context.get('id')
        volume = self.get_volume(volume_id)
        volume.delete()
        logger.info(u'Deleted volume with uuid=%s', volume.id)

    def revert_volume_rename(self, context):
        volume_id = context.get('id')
        old_name = context.get('old_name')
        new_name = context.get('new_name')

        volume = self.get_volume(volume_id)
        volume.update(name=old_name)
        logger.info(
            u'Renamed volume with uuid=%s from %s to %s',
            volume.id, new_name, old_name)

    def revert_volume_detach(self, context):
        server_id = context.get('from_server')
        volume_id = context.get('id')
        device = context.get('device')
        self.nova_client.volumes.create_server_volume(
            server_id, volume_id, device=device)
        logger.info(
            u'Attached volume with uuid=%s to instance with uuid=%s. '
            u'Mount point: %s',
            volume_id, server_id, device)

    def revert_volume_attach(self, context):
        server_id = context.get('to_server')

        server = self.get_server(server_id)
        volume_id = context.get('id')
        volume = self.get_volume(volume_id)
        attachment_id = None

        for attachment in volume.attachments:
            if attachment['server_id'] == server.id:
                attachment_id = attachment['id']
                break

        if attachment_id:
            self.nova_client.volumes.delete_server_volume(
                server.id, attachment_id)
            logger.info(
                u'Detached volume with uuid=%s from instance with uuid=%s',
                volume_id, server_id)

    def revert_port_detach(self, context):
        port_body = context.get('port_info')
        server_id = context.get('from_server')
        server = self.get_server(server_id)
        port = self.neutron_client.create_port(body={'port': port_body})
        server.interface_attach(
            port_id=port['port']['id'], net_id=None, fixed_ip=None)
        logger.info(
            u'Attached port with uuid=%s to instance with uuid=%s.',
            port['port']['id'], server_id)

    def revert_port_create(self, context):
        port_id = context.get('id')
        self.neutron_client.delete_port(port_id)
        logger.info(u'Deleted port with uuid=%s.', port_id)

    def revert_server_create(self, context):
        server_id = context.get('id')
        server = self.get_server(server_id)
        server.delete()
        logger.info(u'Deleted server with uuid=%s.', server_id)
        self.wait_server_deletion(server.id)

    def revert_server_rename(self, context):
        server_id = context.get('id')
        old_name = context.get('old_name')
        server = self.get_server(server_id)
        server.update(name=old_name)
        logger.info(
            u'Renamed server with uuid=%s from %s to %s.',
            server.id, server.name, old_name)


def setup_logging(filepath, level=logging.INFO):
    warnings.filterwarnings("ignore")
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': u'%(asctime)s %(levelname)s %(name)s %(message)s'
            },
            'short': {
                'format': u'%(message)s'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'short',
            },
            'tmp_file': {
                'class': 'logging.FileHandler',
                'formatter': 'default',
                'filename': filepath,
            },
        },
        'root': {
            'handlers': ['console', 'tmp_file'],
            'level': level,
        },
    })


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--os-username',
        default=os.environ.get('OS_USERNAME'))
    parser.add_argument(
        '--os-user-domain-name',
        default=os.environ.get('OS_USER_DOMAIN_NAME'))
    parser.add_argument(
        '--os-password',
        default=os.environ.get('OS_PASSWORD'))
    parser.add_argument(
        '--os-project-name',
        default=os.environ.get('OS_PROJECT_NAME'))
    parser.add_argument(
        '--os-project-domain-name',
        default=os.environ.get('OS_PROJECT_DOMAIN_NAME'))
    parser.add_argument(
        '--os-auth-url',
        default=os.environ.get('OS_AUTH_URL'))
    parser.add_argument(
        '--os-auth-version',
        default=os.environ.get('OS_AUTH_VERSION'))
    parser.add_argument(
        '--db-host',
        default=os.environ.get('DB_HOST'))
    parser.add_argument(
        '--db-port',
        type=int,
        default=os.environ.get('DB_PORT'))
    parser.add_argument(
        '--db-name',
        default=os.environ.get('DB_NAME'))
    parser.add_argument(
        '--db-username',
        default=os.environ.get('DB_USERNAME'))
    parser.add_argument(
        '--db-password',
        default=os.environ.get('DB_PASSWORD'))

    parser.add_argument(
        'server',
        metavar='<server>',
        help='Server ID',
    )
    parser.add_argument(
        '--volume',
        metavar='<volume_uuid:dest_volume_type>',
        action='append',
        default=[],
        help=(
            'Map of block devices to change type; map is '
            '<id>:<new_volume_type>'))
    parser.add_argument(
        '--root-volume-type',
        metavar='<root-volume-type>',
        help='Type of the destination root volume')
    parser.add_argument(
        '--debug',
        action='store_true')
    parser.add_argument(
        '--log',
        action='store_true',
        help='Path to log file')

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--retry',
        action='store_true',
        help='Retry incomplete server migration'
    )
    group.add_argument(
        '--force-complete',
        action='store_true',
        help='Mark incomplete migration as completed'
    )
    group.add_argument(
        '--revert',
        action='store_true',
        help='Mark incomplete migration as completed'
    )
    return parser.parse_args()


def main():
    args = parse_args()

    loglevel = logging.DEBUG if args.debug else logging.INFO
    if args.log:
        filepath = args.log
    else:
        filepath = os.path.join(
            tempfile.gettempdir(), 'tnx_migrate_{0}.log'.format(args.server))

    setup_logging(filepath, loglevel)

    try:
        with MigrationsManager(args.db_host, args.db_port, args.db_name,
                               args.db_username, args.db_password) as manager:
            if args.force_complete:
                migration = manager.get_one(server_id=args.server)
                if not migration:
                    raise CommandError(
                        u'Failed to find migration for server {0}'.format(
                            args.server))
                if migration['status'] == MigrationStatus.REVERTED:
                    raise CommandError(
                        u'Unable to change migration status '
                        u'from {0} to {1}.'.format(
                            MigrationStatus.REVERTED,
                            MigrationStatus.COMPLETED))
                migration['status'] = MigrationStatus.COMPLETED
                manager.update(migration)
            elif args.revert:
                RevertRunner(manager, args).run()
            else:
                MigrationRunner(manager, args).run()
    except CommandError as err:
        logger.error(err, exc_info=args.debug)
        exit(os.EX_SOFTWARE)
    except OperationalError as err:
        logger.error(u'Database error: %s', err, exc_info=args.debug)
        exit(os.EX_SOFTWARE)


if __name__ == '__main__':
    main()
