# coding: utf-8

"""
Base manager working with database
"""

from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from tionix_client.core import exceptions
from tionix_client.core.helpers import get_session


class BaseDBManager(object):
    """
    Base manager working with database
    """

    model = None

    exclude_update_keys = ('id', )
    exclude_create_keys = ('id', )

    default_order_field = 'id'

    @classmethod
    def get_query(cls):
        """
        :return: sqlalchemy query
        """
        session = get_session()
        query = session.query(cls.model)
        return query

    @classmethod
    def count(cls):
        """
        :return: count objects in table
        :rtype: int
        """
        return cls.get_query().count()

    @classmethod
    def create(cls, instance=None, **kwargs):
        """
        :param kwargs: new object params
        :param instance: object for create

        :return: new created object
        :rtype: cls.model
        """
        if instance is not None:
            cls.validate_instance(instance)
        else:
            instance = cls.model()  # pylint: disable=not-callable

        for key, value in kwargs.iteritems():
            if key not in cls.exclude_create_keys and hasattr(instance, key):
                setattr(instance, key, value)

        return cls.save(instance)

    @classmethod
    def create_all(cls, instances):
        """
        Create all instances.
        :param instances: list of db model instances
        :return: list of created instances
        """
        return [cls.create(instance) for instance in instances]

    @classmethod
    def delete(cls, instance_or_id):
        """
        :param instance_or_id: deleted instance object or row id
        :type instance_or_id: cls.model or int
        """
        row_id = getattr(instance_or_id, 'id', instance_or_id)
        cls.get_query().filter_by(id=row_id).delete()

    @classmethod
    def delete_all(cls, instances):
        """
        :param instances: deleted instances object
        :type instances: list of cls.model
        """
        for instance in instances:
            cls.delete(instance)

    @classmethod
    def drop(cls):
        """
        delete all data in table
        """
        return cls.get_query().delete()

    @classmethod
    def get_one(cls, **kwargs):
        """
        :param kwargs: object attributes

        :return: find one object with attributes and return

        :raises exceptions.NoResultFound:
        :raises exceptions.MultipleResultsFound:
        """
        try:
            return cls.get_query().filter_by(**kwargs).one()
        except NoResultFound as err:
            raise exceptions.NoResultFound(unicode(err))
        except MultipleResultsFound as err:
            raise exceptions.MultipleResultsFound(unicode(err))

    @classmethod
    def get_many(cls, **kwargs):
        """
        :param kwargs: object attributes
            - exclude_deleted
            - limit
            - offset

        :return: find all object with attributes and return
        :rtype: list
        """
        limit = None
        offset = None

        query = cls.get_query()

        if 'limit' in kwargs:
            limit = kwargs.pop('limit')

        if 'offset' in kwargs:
            offset = kwargs.pop('offset')

        for k, v in kwargs.items():
            if k.endswith('__in') and isinstance(v, (tuple, list)):
                column_name = k[:-4]
                column = getattr(cls.model, column_name, None)
                if column:
                    kwargs.pop(k)
                    query = query.filter(column.in_(v))

        query = query.filter_by(**kwargs).order_by(cls.default_order_field)

        if limit:
            query = query.limit(limit)

        if offset:
            query = query.offset(offset)

        return query.all()

    @classmethod
    def _save(cls, instance):
        """
        :param instance: object for save

        :return: saved object
        """
        # Fixing the AssertionError
        fields = instance.__dict__.keys()
        for r_field in cls.model.relation_filed:
            if r_field in fields and r_field + '_id' in fields:
                instance.__dict__.pop(r_field)

        session = get_session()
        session.add(instance)
        session.flush()

        return instance

    @classmethod
    def save(cls, instance):
        """
        :param instance: object for save

        :raises ValueError:

        :return: saved object
        """
        cls.validate_instance(instance)
        cls._save(instance)
        return instance

    @classmethod
    def update_one(cls, instance, **kwargs):
        """
        :param instance: instance or instance_id for update
        :type instance: cls.model or int

        :param kwargs: instance update attributes

        :raises ValueError:

        :return: updated instance
        :rtype: cls.model
        """
        instance = cls.get_instance(instance)

        for key in kwargs:
            if key not in cls.exclude_update_keys and hasattr(instance, key):
                setattr(instance, key, kwargs[key])

        return cls.save(instance)

    @classmethod
    def update_many(cls, instances):
        """
        :param instances: instances for update
        :type instances: list

        :raises ValueError:
        """
        for instance in instances:
            cls.validate_instance(instance)

        session = get_session()

        if hasattr(session, 'bulk_save_objects'):
            session.bulk_save_objects(instances)
        else:
            session.add_all(instances)

        session.flush()

    @classmethod
    def validate_instance(cls, instance):
        """Override this method to add fields validation"""
        pass

    @classmethod
    def create_instance(cls, **kwargs):
        """Creates valid model instance without actually storing it in DB."""
        for k in cls.exclude_create_keys:
            kwargs.pop(k, None)
        instance = cls.model(**kwargs)  # pylint: disable=not-callable
        cls.validate_instance(instance)
        return instance

    @classmethod
    def get_instance(cls, instance):
        if not isinstance(instance, cls.model):
            instance = cls.get_one(id=instance)

        return instance
