# coding: utf-8

from __future__ import absolute_import

import datetime
import json
import logging
import six

from nova import config, context as nova_context, objects
from nova.db.sqlalchemy import utils as db_utils
from oslo_db import exception as db_exc
from oslo_utils import timeutils
from sqlalchemy import MetaData, sql, desc
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.schema import Table

from tionix_client.journal.constants import ObjectType, ActionStatus
from tionix_client.journal.manager import JournalManager
from tionix_client.os_clients.nova_client import get_admin_nova_client

logger = logging.getLogger(__name__)

SHADOW_TABLE_PREFIX = 'shadow_'


# noinspection PyUnusedLocal
def archive_incorrect_rows_for_table(tablename, instances, connection=None,
                                     metadata=None, base=None, **kwargs):
    table = base.metadata.tables.get(tablename)

    if table is None:
        logger.error('Metadata for table %s is not found.', tablename)
        return 0

    shadow_tablename = SHADOW_TABLE_PREFIX + tablename
    rows_archived = 0
    try:
        shadow_table = Table(shadow_tablename, metadata, autoload=True)
    except NoSuchTableError:
        return rows_archived

    try:
        instance_id = table.c.instance_id
    except AttributeError:
        logger.debug("Table %s has no reference to instances table.",
                     tablename)
        return rows_archived

    column = table.c.id  # primary key
    columns = [c.name for c in table.c]

    # pylint: disable=no-value-for-parameter
    insert = shadow_table.insert(inline=True).from_select(
        columns, sql.select(
            [table], ~instance_id.in_(instances)).order_by(column))
    # pylint: disable=no-value-for-parameter
    query_delete = sql.select(
        [column], ~instance_id.in_(instances)).order_by(column)

    delete_statement = db_utils.DeleteFromSelect(table, query_delete, column)
    try:
        # Group the insert and delete in a transaction.
        with connection.begin():
            connection.execute(insert)
            result_delete = connection.execute(delete_statement)
        rows_archived = result_delete.rowcount
    except db_exc.DBReferenceError as ex:
        # A foreign key constraint keeps us from deleting some of
        # these rows until we clean up a dependent table.  Just
        # skip this table for now; we'll come back to it later.
        logger.warning("IntegrityError detected when archiving table "
                       "%(tablename)s: %(error)s",
                       {'tablename': tablename, 'error': six.text_type(ex)})
    return rows_archived


def _confirm(msg, std):
    stdin, stdout = std
    while True:
        stdout.write(msg)
        stdout.flush()
        answer = stdin.readline().strip()

        if answer.lower() in ('y', 'n', 'yes', 'no'):
            break
    return answer in ('y', 'yes')


# noinspection PyUnusedLocal
def archive_aged_rows_for_table(age_in_days, connection=None, metadata=None,
                                std=None, conf=None, base=None, **kwargs):
    tablename = 'sessions'
    sessions_table = base.metadata.tables.get(tablename)
    if sessions_table is None:
        logger.error('Metadata for table %s is not found.', tablename)
        return

    expired_age = timeutils.utcnow() - datetime.timedelta(days=age_in_days)

    deleted_column = sessions_table.c.deleted
    instances = set()
    try:
        with connection.begin():
            query = sql.select(
                [sessions_table],
                deleted_column == deleted_column.default.arg
            ).order_by(desc(sessions_table.c.id))
            sessions = connection.execute(query)

            exclude = set()
            for session in sessions:
                if session.instance_id in exclude:
                    continue
                if session.created_at and session.created_at < expired_age:
                    instances.add(session.instance_id)
                else:
                    exclude.add(session.instance_id)
    except Exception as err:
        logger.error(unicode(err))

    instances_count = len(instances)
    if instances_count:
        confirm = _confirm("%s instances will be removed. "
                           "Are you sure? (y/n)" % instances_count, std)
        if confirm:
            nova_client = get_admin_nova_client(conf)
            for instance_id in instances:
                try:
                    logger.debug('Instance with id %s is removing.',
                                 instance_id)
                    nova_client.servers.delete(instance_id)
                except Exception as err:
                    logger.error(unicode(err))


# noinspection PyUnusedLocal
def archive_deleted_rows_for_table(tablename, connection=None, metadata=None,
                                   base=None, **kwargs):
    table = base.metadata.tables.get(tablename)

    if table is None:
        logger.error('Metadata for table %s is not found.', tablename)
        return 0

    shadow_tablename = SHADOW_TABLE_PREFIX + tablename
    rows_archived = 0
    try:
        shadow_table = Table(shadow_tablename, metadata, autoload=True)
    except NoSuchTableError:
        return rows_archived

    column = table.c.id  # primary key
    deleted_column = table.c.deleted
    columns = [c for c in table.c if c.name != column.name]

    # pylint: disable=no-value-for-parameter
    insert = shadow_table.insert(inline=True).from_select(
        columns, sql.select(
            columns,
            deleted_column != deleted_column.default.arg).order_by(column))
    query_delete = sql.select(
        [column],
        deleted_column != deleted_column.default.arg).order_by(column)

    delete_statement = db_utils.DeleteFromSelect(table, query_delete, column)
    try:
        # Group the insert and delete in a transaction.
        with connection.begin():
            connection.execute(insert)
            result_delete = connection.execute(delete_statement)
        rows_archived = result_delete.rowcount
    except db_exc.DBReferenceError as ex:
        # A foreign key constraint keeps us from deleting some of
        # these rows until we clean up a dependent table.  Just
        # skip this table for now; we'll come back to it later.
        logger.warning("IntegrityError detected when archiving table "
                       "%(tablename)s: %(error)s",
                       {'tablename': tablename, 'error': six.text_type(ex)})
    return rows_archived


def archive_rows_for_journal(tablename, age=None, connection=None,
                             metadata=None):
    table = metadata.tables.get(tablename)
    shadow_tablename = SHADOW_TABLE_PREFIX + tablename

    rows_archived = 0
    try:
        shadow_table = Table(shadow_tablename, metadata, autoload=True)
    except NoSuchTableError:
        return rows_archived

    column = table.c.id  # primary key
    columns = [c for c in table.c if c.name != column.name]
    expired_age = timeutils.utcnow() - datetime.timedelta(days=age)
    event_time = table.c.event_time

    delete_query = sql.select([column], event_time <= expired_age).order_by(
        column)
    # pylint: disable=no-value-for-parameter
    insert = shadow_table.insert(inline=True).from_select(
        columns, sql.select(columns, event_time <= expired_age).order_by(
            column))
    delete_statement = db_utils.DeleteFromSelect(table, delete_query, column)

    try:
        with connection.begin():
            connection.execute(insert)
            result_delete = connection.execute(delete_statement)
        rows_archived = result_delete.rowcount
    except Exception as err:
        logger.error(u"Couldn't archive journal table: {0}".format(err))
    return rows_archived


def _purge(connection, incorrect=None, age=None, std=None, **kwargs):
    rows_archived = 0
    metadata = MetaData(connection)
    metadata.reflect()
    affected_tables = []

    if age:
        archive_aged_rows_for_table(age, connection=connection,
                                    metadata=metadata, std=std, **kwargs)
        return rows_archived, affected_tables

    instances = set()
    if incorrect:
        config.parse_args([''])
        objects.register_all()
        context = nova_context.get_admin_context()
        manager = objects.InstanceList  # pylint: disable=no-member
        # get all instances excluding deleted
        instances = {instance.uuid for instance in manager.get_all(context)}

    # Reverse sort the tables so we get the leaf nodes first for processing.
    for table in reversed(metadata.sorted_tables):
        table_name = table.name
        exclude = ('migrate_version', 'celery_taskmeta', 'celery_tasksetmeta',
                   'journal')
        if (table_name in exclude or
                table_name.startswith(SHADOW_TABLE_PREFIX)):
            continue

        affected_rows = 0
        if incorrect:
            affected_rows += archive_incorrect_rows_for_table(
                table_name, instances, connection=connection,
                metadata=metadata, **kwargs)
        affected_rows += archive_deleted_rows_for_table(
            table_name, connection=connection, metadata=metadata, **kwargs)

        if affected_rows:
            affected_tables.append(table_name)

        rows_archived += affected_rows
    return rows_archived, affected_tables


def purge(connection, incorrect=None, age=None, std=None, **kwargs):
    func_kwargs = {
        'incorrect': incorrect,
        'age': age,
    }

    object_id = user_id = kwargs.get('user_id')
    user_name = kwargs.get('user_name')
    domain_id = kwargs.get('domain_id')
    object_type = ObjectType.User
    module = connection.engine.url.database
    action = 'openstack.tnx.db.purge.' + module
    action_arguments = json.dumps(func_kwargs)

    status = ActionStatus.SUCCESS
    message = ''
    try:
        rows, tables = _purge(connection, incorrect=incorrect, age=age,
                              std=std, **kwargs)
        if tables:
            message = 'Affected tables: {}.'.format(', '.join(tables))
        else:
            message = 'No affected tables.'
    except Exception as err:
        status = ActionStatus.FAILURE
        message = unicode(err)
        raise
    finally:
        JournalManager.log_action(object_id, object_type, action,
                                  is_notification=False,
                                  event_time=datetime.datetime.utcnow(),
                                  user_id=user_id,
                                  user_name=user_name,
                                  action_arguments=action_arguments,
                                  status=status,
                                  message=message)
    return rows


def purge_journal(connection, age):
    metadata = MetaData(connection)
    metadata.reflect()
    tablename = 'journal'

    return archive_rows_for_journal(
        tablename=tablename, connection=connection, metadata=metadata, age=age)
