# coding: utf-8

from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import NoSuchTableError, SQLAlchemyError
from oslo_db.exception import DBError

from tionix_client.settings import get_db_connection_url, CONF


table_name = 'journal'

MODULE_DBS = (
    'tionix_cloud_control', 'tionix_monitor', 'tionix_node_control',
    'tionix_vdi_server')


def transform_row(row):
    result = {
        'is_notification': False
    }
    source_columns = (
        'created_at', 'request_id', 'parent_request_id', 'object_type',
        'object_id', 'action', 'user_name', 'user_id', 'status', 'result')
    dest_columns = {
        'created_at': 'event_time',
        'result': 'message'}
    formatters = {
        'status': lambda s: s.lower() if s else 'unknown',
        'object_type': lambda t: 'node' if t == 'hypervisor' else t}

    for column in source_columns:
        val = getattr(row, column, None)
        formatter = formatters.get(column)
        if formatter:
            val = formatter(val)
        new_column = dest_columns.get(column, column)
        result[new_column] = val
    return result


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    table = Table(table_name, meta, autoload=True)
    # pylint:disable=no-value-for-parameter
    insert = table.insert()

    for module_db in MODULE_DBS:
        module_conn_str = get_db_connection_url(CONF, db_name=module_db)
        try:
            module_engine = create_engine(
                module_conn_str, echo=False, convert_unicode=True)
        except (SQLAlchemyError, DBError):
            continue

        module_session = sessionmaker(bind=module_engine)()
        module_meta = MetaData(bind=module_engine)
        try:
            module_table = Table(table_name, module_meta, autoload=True)
        except NoSuchTableError:
            continue

        delete_ids = []
        for row in module_session.query(module_table).filter(
                module_table.c.deleted == 0).all():
            new_row = transform_row(row)
            try:
                insert.execute(new_row)
            except (SQLAlchemyError, DBError):
                # skip invalid rows
                continue
            else:
                delete_ids.append(row.id)

        if delete_ids:
            module_table.delete().where(
                module_table.c.id.in_(delete_ids)).execute()


# noinspection PyUnusedLocal
def downgrade(migrate_engine):
    pass
