# coding: utf-8

import datetime
import logging
import inspect

from tionix_client.helpers import packages_info
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.helpers.strutils import mask_dict_password
from tionix_client.helpers.table import build_wrapped_table
from tionix_client.helpers.updates import get_latest_package_version

logger = logging.getLogger(__name__)


STATUS_TRANSLATIONS = _(u'OK'), _(u'ERROR'), _(u'FAIL'), _(u'SKIPPED')
HIDDEN_PASSWORD = "********"
UNSAFE_CHUNK = 'pass'


class TestResultStatus(object):
    OK = u'OK'
    ERROR = u'ERROR'
    FAIL = u'FAIL'
    SKIPPED = u'SKIPPED'

    map = {
        OK: u'success',
        ERROR: u'errors',
        FAIL: u'failures',
        SKIPPED: u'skipped',
    }


class ModuleNames(object):
    tnx_cloud_control = u'TIONIX.CloudControl'
    tnx_dashboard = u'TIONIX.Dashboard'
    tnx_monitor = u'TIONIX.Monitor'
    tnx_node_control = u'TIONIX.NodeControl'
    tnx_scheduler = u'TIONIX.Scheduler'
    tnx_scheduler_client = u'TIONIX.SchedulerClient'
    tnx_client = u'TIONIX.Client'
    tnx_vdi_server = u'TIONIX.VDIserver'
    tnx_vdi_client = u'TIONIX.VDIclient'

modules_info_getters = {
    ModuleNames.tnx_cloud_control: packages_info.get_cloud_control_info,
    ModuleNames.tnx_node_control: packages_info.get_node_control_info,
    ModuleNames.tnx_monitor: packages_info.get_monitor_info,
    ModuleNames.tnx_vdi_server: packages_info.get_vdi_server_info,
    ModuleNames.tnx_dashboard: packages_info.get_dashboard_info,
    ModuleNames.tnx_scheduler_client: packages_info.get_scheduler_client_info,
    ModuleNames.tnx_client: packages_info.get_tionix_client_info,
    ModuleNames.tnx_scheduler: packages_info.get_scheduler_info,
    ModuleNames.tnx_vdi_client: packages_info.get_tionix_vdi_client_info,
}

licensed_modules = (
    ModuleNames.tnx_cloud_control,
    ModuleNames.tnx_node_control,
    ModuleNames.tnx_monitor,
    ModuleNames.tnx_vdi_server,
)


def module_info_provider(module_name, nova_client):
    result = (None, u'-', '', None)
    no_license_msg = _(u'This module does not require a license.')

    info_func = modules_info_getters.get(module_name)
    if info_func:

        func_args = inspect.getargspec(info_func).args
        kwargs = {}
        if 'nova_client' in func_args:
            kwargs['nova_client'] = nova_client

        info = info_func(**kwargs)

        if info:
            latest_version = get_latest_package_version(info.pip_name)
            result = (
                info.version,
                info.lic_id if info.lic_id != u'-' else no_license_msg,
                info.expire_at or '',
                latest_version)
    return result


def format_module_name(module_name):
    return getattr(
        ModuleNames,
        module_name,
        module_name.replace('_', ' ').capitalize()
    )


def add_license_checking_result(test_result, module_name, nova_client):
    info_func = modules_info_getters.get(module_name)
    if info_func and module_name in licensed_modules:
        test_name = u'test license validity'
        result = test_result['tests']
        module_info = info_func(nova_client)

        license_verified = module_info.verified
        license_exp_date = module_info.expire_at
        err = warn = ''

        if license_verified is None:
            err = _(u'License is not installed.')
        elif license_verified:
            if license_exp_date != '-':
                d1 = datetime.datetime.strptime(license_exp_date,
                                                '%d.%m.%Y %H:%M:%S')
                d2 = datetime.datetime.now()
                days_delta = abs((d2 - d1).days)
                if days_delta <= 30:
                    warn = (_(u'WARNING! License will expire in %d days')
                            % days_delta)
        else:
            err = _(u'License has expired.')

        if not nova_client:
            result['skipped'].append(
                (test_name, _(u"Couldn't connect to OpenStack")))
        elif warn:
            result['success'].append((test_name, warn))
        elif err:
            result['failures'].append((test_name, err))
        else:
            result['success'].append((test_name, ''))


def config_diff(module_conf, general_conf, prefix=''):
    diff_config = {}
    for key, value in module_conf.items():
        general_value = general_conf.get(key, '') if isinstance(
            general_conf, dict) else ''
        if isinstance(value, dict):
            diff_config.update(
                config_diff(value, general_value, prefix=key + "."))
        else:
            result_key = prefix + key
            if general_value != value:
                diff_config[result_key] = (
                    HIDDEN_PASSWORD
                    if UNSAFE_CHUNK in result_key.lower() else value
                )
    return diff_config


def build_variables_table(module_name, module_conf, general_conf):
    result = ''
    if module_conf:
        variables = mask_dict_password(config_diff(module_conf, general_conf))
        headers = [u'N', u'Variable name', u'Value']
        rows = []
        for i, (variable, value) in enumerate(
                iter(sorted(variables.items())), 1):
            if isinstance(value, (list, tuple)):
                value = u", ".join(value)
            rows.append([i, variable, value])

        formatted_table = build_wrapped_table(headers, rows)

        table_name = _(u'{} extracted variables').format(module_name)
        table_data = (
            u"{border}\n" u"{table_name}\n" u"{border}\n" u"{table_data}"
        ).format(table_name=table_name, border="=" * len(table_name),
                 table_data=formatted_table + '\n')
        result = "\n" + table_data
    return result
