# coding: utf-8

import fcntl
import os


class LockError(Exception):
    """Error while setting lock"""


class FileLocker(object):
    """FileLocker implements advisory lock on file

    Lock should be applied after opening file for writing. Advisory lock
    doesnt't prevent file from read/write operations, to check if file is
    locked use 'is_locked' method
    """

    file_path = None
    _fh = None

    def __init__(self, file_path):
        super(FileLocker, self).__init__()
        self.file_path = file_path

    @property
    def fh(self):
        """Get opened file handler"""

        if self._fh is None:
            try:
                self._fh = open(self.file_path, 'w')
            except (IOError, OSError) as err:
                raise LockError(unicode(err))
        return self._fh

    @staticmethod
    def _lock(fh):
        try:
            fcntl.flock(fh, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except (IOError, ValueError) as err:
            raise LockError(unicode(err))

    def lock(self):
        """Apply lock on file"""
        self._lock(self.fh)

    @staticmethod
    def _unlock(fh):
        try:
            fcntl.flock(fh, fcntl.LOCK_UN)
        except IOError as err:
            raise LockError(unicode(err))

    def unlock(self):
        """Remove lock from file"""

        try:
            self._unlock(self.fh)
        except LockError:
            pass
        else:
            self._fh.close()
            self._fh = None

    def is_locked(self):
        """Check if file is locked"""

        is_locked = False
        # it is important to get new file handler for the same file to
        # check if it was locked, do not close original fh because lock will
        # be released
        if os.path.isfile(self.file_path):
            with open(self.file_path) as fh:
                try:
                    self._lock(fh)
                except LockError:
                    is_locked = True
                else:
                    self._unlock(fh)
        return is_locked
