# coding: utf-8
import datetime

from tionix_client.functional.helpers import (
    TestResultStatus, module_info_provider, add_license_checking_result,
    build_variables_table)
from tionix_client.functional.settings import get_general_config
from tionix_client.helpers.table import build_wrapped_table
from tionix_client.helpers.i18 import ugettext as _


class Result(object):
    def __init__(self, test_result=None):
        if test_result:
            if not isinstance(test_result, dict):
                raise TypeError(
                    u"Cannot create '{0}' from '{1}' object".format(
                        type(self).__name__, type(test_result).__name__)
                )
            self.run = sum(map(len, test_result['tests'].values()))
            self.error = len(test_result['tests']['errors'])
            self.fail = len(test_result['tests']['failures'])
            self.skipped = len(test_result['tests']['skipped'])
            self.success = len(test_result['tests']['success'])
            self.time = test_result['time']
            self.was_successful = test_result['was_successful']
        else:
            self.run = 0
            self.success = 0
            self.error = 0
            self.skipped = 0
            self.fail = 0
            self.time = 0
            self.was_successful = True

    def __add__(self, other):
        if not isinstance(other, Result):
            raise TypeError(
                u"Cannot concatenate '{0}' and '{1}' objects".format(
                    type(self).__name__, type(other).__name__)
            )
        result = Result()
        result.run = self.run + other.run
        result.success = self.success + other.success
        result.error = self.error + other.error
        result.fail = self.fail + other.fail
        result.skipped = self.skipped + other.skipped
        result.time = self.time + other.time
        result.was_successful = self.was_successful & other.was_successful
        return result


class TestParser(object):
    """Class for parsing DetailedTestResult"""

    def __init__(self):
        self.total_result = Result()
        self.current_result = Result()
        self.module_info = None
        self.module_test_result = None
        self.variables_table = ''
        self.general_conf = get_general_config()

    def _update_stats(self, test_result):
        if test_result:
            self.current_result = Result(test_result)
            self.total_result += self.current_result

    def get_module_summary(self, errors_only=False):
        """Get module tests summary"""
        table = self._make_table(
            self.module_test_result, errors_only=errors_only)

        detailed_status = _(
            u"success={success}, failures={fails}, "
            u"errors={errors})").format(
                success=self.current_result.success,
                fails=self.current_result.fail,
                errors=self.current_result.error)

        summary = _(
            u"{module_info}"
            u"{table}"
            u"Ran {run} in {time:.3f}s\n"
            u"{status} ({detailed_status})\n\n"
        ).format(
            module_info=self.module_info,
            table=table,
            run=self.current_result.run,
            time=self.current_result.time,
            status=(
                _(TestResultStatus.OK) if self.current_result.was_successful
                else _(TestResultStatus.FAIL)),
            detailed_status=detailed_status)

        return summary

    def get_summary(self):
        """Get all modules tests summary"""
        summary = _(
            u"======\n"
            u"TOTALS\n"
            u"======\n"
            u"Ran {total_run} in {time:.3f}s\n"
            u"{status} (success={success}, "
            u"failures={fails}, errors={errors})\n"
        ).format(
            total_run=self.total_result.run,
            time=self.total_result.time,
            status=(
                _(TestResultStatus.OK) if self.total_result.was_successful
                else _(TestResultStatus.FAIL)),
            success=self.total_result.success,
            fails=self.total_result.fail,
            errors=self.total_result.error
        )

        return summary

    def _make_table(self, test_result, errors_only=False):
        formatted_table = u''
        tests_list = [
            (test_result['tests']['failures'], TestResultStatus.FAIL),
            (test_result['tests']['errors'], TestResultStatus.ERROR),
        ]

        if not errors_only:
            tests_list.extend([
                (test_result['tests']['success'], TestResultStatus.OK),
                (test_result['tests']['skipped'], TestResultStatus.SKIPPED),
            ])

        headers = [u'N', _(u'Test name'), _(u'Status'), _(u'Failure reason')]
        rows = []
        for tests, status in tests_list:
            rows.extend(
                self._parse_tests(
                    tests=tests,
                    status=_(status)
                )
            )
        rows = sorted(rows, key=lambda x: x[1])
        for i, row in enumerate(rows):
            row[0] = i + 1

        if rows:
            formatted_table = build_wrapped_table(headers, rows) + '\n'

        return formatted_table

    @staticmethod
    def _parse_tests(tests, status):
        rows = []
        for test, reason in tests:
            rows.append(
                [None, test, status, reason]
            )
        return rows

    def get_module_variables(self):
        return self.variables_table

    def process_result(self, test_result, module_name, nova_client=None):
        """
        Process dict obtained from DetailedTestResult.to_dict method
        """
        test_modules = [
            'Fake module',
            'Test'
        ]
        installed_version = None
        latest_available_version = None
        lic = u'-'
        lic_valid_date = ''
        expire_at = ''

        if module_name not in test_modules:
            module_info = module_info_provider(module_name, nova_client)
            (installed_version, lic, expire_at,
             latest_available_version) = module_info

        if latest_available_version == installed_version:
            latest_available_version = _(u'current')

        if isinstance(test_result, dict):
            add_license_checking_result(test_result, module_name, nova_client)
            module_variables = test_result.get('variables')
            self.variables_table = build_variables_table(
                module_name, module_variables, self.general_conf.items())
            self.module_test_result = test_result
            table_name = module_name
            self._update_stats(test_result)
            if expire_at and expire_at != '-':
                formatted_date = datetime.datetime.strptime(
                    expire_at, '%d.%m.%Y %H:%M:%S')
                lic_valid_date = _(
                    '(valid until {formatted_date:%d.%m.%Y})').format(
                        formatted_date=formatted_date)
            self.module_info = _(
                u"{border}\n"
                u"{table_name}\n"
                u"Version: {version} (latest: {latest_version})\n"
                u"License: {lic} {lic_valid_date} \n"
                u"{border}\n"
            ).format(
                table_name=table_name,
                version=installed_version,
                latest_version=latest_available_version or _(u'N\\A'),
                lic=lic,
                lic_valid_date=lic_valid_date,
                border="=" * len(table_name)
            )
