# coding: utf-8

import time
import traceback

from collections import defaultdict
from StringIO import StringIO
from unittest import TestResult, TextTestRunner, case

from tionix_client.functional.helpers import TestResultStatus


class _WritelnDecorator(object):
    """Used to decorate file-like objects with a handy 'writeln' method"""
    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, attr):
        if attr in ('stream', '__getstate__'):
            raise AttributeError(attr)
        return getattr(self.stream, attr)

    def writeln(self, arg=None):
        if arg:
            self.write(arg)
        self.write('\n')  # text-mode streams translate to \r\n if needed


class SkipTest(case.SkipTest):
    def __init__(self, reason=None):
        # convert unicode string to bytestring to prevent
        # UnicodeEncodeError in unittest.TestCase
        if isinstance(reason, unicode):
            reason = reason.encode('utf-8')
        super(SkipTest, self).__init__(reason)


class DetailedTestResult(TestResult):
    """ Extended TestResult subclass

    DetailedTestResult is used by DetailedTestRunner to provide additional
    tests information.
    """
    _original_skip_test = None

    def __init__(self, stream, descriptions, verbosity, debug):
        super(DetailedTestResult, self).__init__(
            stream, descriptions, verbosity)
        self.debug = debug
        self.success = []
        self._start_time = 0
        self.time = 0
        self.results = {
            'tests': defaultdict(list),
            'was_successful': True,
            'tests_run': 0,
            'time': 0
        }

    def addSuccess(self, test):
        self.success.append(test)
        self.process_test(test, status=TestResultStatus.OK)

    def addError(self, test, err):
        super(DetailedTestResult, self).addError(test, err)
        if 'ModuleImportFailure' in test.__class__.__name__:
            exctype, value, tb = err
            value = value.message.splitlines()[0]
            err = (exctype, value, tb)
        self.process_test(
            test,
            status=TestResultStatus.ERROR,
            details=self._exc_info_to_string(err, test))

    def addFailure(self, test, err):
        super(DetailedTestResult, self).addFailure(test, err)
        self.process_test(
            test,
            status=TestResultStatus.FAIL,
            details=self._exc_info_to_string(err, test))

    def addSkip(self, test, reason):
        super(DetailedTestResult, self).addSkip(test, reason)
        self.process_test(
            test, status=TestResultStatus.SKIPPED, details=reason)

    def _exc_info_to_string(self, err, test):
        exctype, value, tb = err
        if exctype is AssertionError:
            message = value.message
            if isinstance(message, str):
                message = message.decode('utf8')
        else:
            if not self.debug:
                tb = None
            message = u''.join(
                traceback.format_exception(exctype, value, tb))
        return message

    def startTestRun(self):
        # patch original unittest SkipTest exception to allow
        # non-ascii symbols in skip test reason
        self._original_skip_test = case.SkipTest
        case.SkipTest = SkipTest
        self._start_time = time.time()

    def stopTestRun(self):
        case.SkipTest = self._original_skip_test
        self.time = time.time() - self._start_time

    def process_test(self, test, status, details=u''):
        """Process test and update results dict"""
        test_name = unicode(test).split()[0].replace('_', ' ')
        if isinstance(details, str):
            details = details.decode('utf-8')
        key = TestResultStatus.map.get(status)
        self.results['tests'][key].append((test_name, details))

    def to_dict(self):
        """returns dict with stored tests information"""
        self.results['was_successful'] = self.wasSuccessful()
        self.results['time'] = self.time
        self.results['tests_run'] = self.testsRun
        return self.results


class DetailedTestRunner(TextTestRunner):
    """A test runner class with suppressed output and extended resultclass"""

    def __init__(self, debug=False, *args, **kwargs):
        super(DetailedTestRunner, self).__init__(*args, **kwargs)
        self.stream = _WritelnDecorator(StringIO())
        self.resultclass = DetailedTestResult
        self.debug = debug

    def _makeResult(self):
        return self.resultclass(self.stream, self.descriptions, self.verbosity,
                                self.debug)

    def get_result(self):
        return self._makeResult()
