# coding: utf-8

import logging
import os
import sys

from tionix_client.functional.constants import (
    REPORT_DIR, REPORT_NAME, REPORT_POSTFIX_LATEST)
from tionix_client.functional.locker import FileLocker, LockError

logger = logging.getLogger(__name__)


class BaseWriter(object):
    """Writer interface"""
    def init_writer(self):
        raise NotImplementedError

    def write(self, message):
        raise NotImplementedError

    def close_writer(self):
        raise NotImplementedError


class TestsRunningException(Exception):
    pass


class VersionedFileWriter(BaseWriter):
    """Writer for versioned files.

    Args:
        file_dir (str): Directory, where report file will be created.
        file_name (str): Base name of report file.
        latest_ext (str): Extension of latest file.
        username (str): User name, will be added to report name.

    """

    def __init__(self, file_dir, file_name, latest_ext, username,
                 locker_class=FileLocker):
        self.file_dir = file_dir or REPORT_DIR
        self.username = username
        self.file_name = u'{0}_{1}'.format(username, file_name)

        self.latest_path = os.path.join(
            self.file_dir, u'{0}.{1}'.format(self.file_name, latest_ext))
        self.locker = locker_class(self.latest_path)

        self._initialized = False
        self._file = None

    def __repr__(self):
        return u"VersionedFileWriter(file_path='{0}')".format(self.latest_path)

    def init_writer(self):
        """Init file

        Changes existing file extension from '.latest' to incrementing number
        and create file.

        """
        msg_running = u'Functional tests are running already.'
        if self.locker.is_locked():
            raise TestsRunningException(msg_running)

        if not os.path.exists(self.file_dir):
            try:
                os.mkdir(self.file_dir)
            except OSError:
                msg = u"Couldn't create report dir:'{0}'.".format(
                    self.file_dir)
                raise TestsRunningException(msg)

        if os.path.isfile(self.latest_path):
            try:
                os.rename(self.latest_path, self.new_version)
            except OSError:
                msg = u"Couldn't rename tests report '{0}' to '{1}'.".format(
                    self.latest_path, self.new_version)
                raise TestsRunningException(msg)

        try:
            self._file = open(self.latest_path, 'w')
        except (IOError, OSError):
            msg = u"Couldn't open file '{0}' for writing.".format(
                self.latest_path)
            raise TestsRunningException(msg)
        try:
            self.locker.lock()
        except LockError:
            raise TestsRunningException(msg_running)

        self._initialized = True

    def write(self, message):
        """Log message to existing file"""
        if self._initialized:
            self._file.write(message)

    def close_writer(self):
        """Unlock and close file"""
        if self._initialized:
            self.locker.unlock()
            self._file.close()

    @property
    def new_version(self):
        """Get new available file name"""
        i = 0
        while True:
            i += 1
            new_name = os.path.join(
                self.file_dir, u'{0}.{1}'.format(self.file_name, i))
            if not os.path.exists(new_name):
                break
        return new_name

    def get_content(self):
        content = None
        if self.locker.is_locked():
            logger.error(
                u'Unable to get latest report, functional tests are running.')
        else:
            try:
                with open(self.latest_path) as fh:
                    content = fh.read()
            except IOError:
                logger.error(
                    u"Couldn't find latest report for user '{0}'".format(
                        self.username))
        return content


class StdOutWriter(BaseWriter):
    """Writer for stdout"""
    def __init__(self, stdout=sys.stdout):
        self.stdout = stdout

    def __repr__(self):
        return u"StdOutWriter(stdout='{0}')".format(self.stdout)

    def init_writer(self):
        pass

    def write(self, message):
        self.stdout.write(message)

    def close_writer(self):
        pass


class SocketWriter(BaseWriter):
    """Writer for socket objects."""
    def __init__(self, socket):
        self.socket = socket
        self.opened = True

    def init_writer(self):
        pass

    def write(self, message):
        if self.opened:
            try:
                self.socket.send_chunk(message)
            except Exception as err:
                logger.error(u"Couldn't write to socket because of the "
                             u"following error: {0}".format(unicode(err)))
                self.opened = False

    def close_writer(self):
        pass


def get_writers(silent=True, save_to_file=False, file_dir=None,
                socket=None, username=None):
    """Get list of writers.

    Args:
        silent (bool): If False, test results will be printed to stdout.
        save_to_file (bool): If True, test results will be logged to file.
        file_dir (str): Directory, where report file will be created.
        socket: socket object with `send_chunk` method.
        username (str): Name of current user.

    Returns:
        list: list of writers

    """
    writers = []
    if not silent:
        writers.append(StdOutWriter(stdout=sys.stdout))
    if save_to_file:
        writers.append(VersionedFileWriter(file_dir=file_dir,
                                           file_name=REPORT_NAME,
                                           latest_ext=REPORT_POSTFIX_LATEST,
                                           username=username))
    if socket:
        writers.append(SocketWriter(socket=socket))
    return writers
