import logging
import sys

from tionix_client.helpers.strutils import mask_dict_password, mask_password
from tionix_client.settings import CONF


class FormatterWithExceptions(logging.Formatter):
    """
    Custom logging formatter that appends traceback to the logger output
     if needed.
    :tb_log_level: set to show tracebacks only for logger levels
     equal or higher.
    """

    def __init__(self, fmt=None, datefmt=None, show_traceback=False,
                 tb_log_level=logging.ERROR):
        super(FormatterWithExceptions, self).__init__(fmt, datefmt)
        self.tb_log_level = tb_log_level
        self.show_traceback = show_traceback

    def format(self, record):
        if self.show_traceback:
            if record.levelno >= self.tb_log_level and not record.exc_info:
                exc_info = sys.exc_info()
                if all(exc_info):  # if has captured exception
                    record.exc_info = exc_info
        return super(FormatterWithExceptions, self).format(record)


class SecretsSanitizerFilter(logging.Filter):
    """Hides sensitive data"""

    def filter(self, record):
        record.msg = mask_password(record.msg)
        if isinstance(record.args, dict):
            record.args = mask_dict_password(record.args)
        else:
            record.args = tuple(mask_password(arg) for arg in record.args)
        return True


def setup_module_logger(module_name, level=None, log_fmt=None,
                        date_format=None, force_tb=CONF.TRACEBACK_ENABLED):
    """Set custom handler that forces to show traceback."""
    logger = logging.getLogger(module_name)
    formatter = FormatterWithExceptions(log_fmt, date_format, force_tb)
    if level:
        logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger.propagate = 0
    logger.addHandler(handler)
