import collections
import functools
from datetime import datetime, timedelta


class InvalidCache(Exception):
    pass


_CachedItem = collections.namedtuple(
    'CachedItem',
    ['value', 'cached_at'])


def lru_cache(maxsize=None, ttl=None):
    def decorating_function(func):
        cache = collections.OrderedDict()

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            key = str(args) + str(kwargs)
            try:
                cached_item = cache.pop(key)
                if ttl and timedelta(seconds=ttl) <= (
                        datetime.utcnow() - cached_item.cached_at):
                    raise InvalidCache

            except (KeyError, InvalidCache):
                val = func(*args, **kwargs)
                if maxsize and len(cache) >= maxsize:
                    cache.popitem(0)
                cache[key] = _CachedItem(
                    value=val, cached_at=datetime.utcnow())
            else:
                val = cached_item.value
                cache[key] = cached_item
            return val

        return wrapper

    return decorating_function


def memoized(func):
    cache = func.cache = {}

    @functools.wraps(func)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    return memoizer
