# coding: utf-8
# pylint: disable=no-name-in-module,import-error

import logging
import os
import subprocess
import codecs

from distutils.version import LooseVersion

import requests

from lxml import html

from tionix_client.helpers import PACKAGES
from tionix_client.helpers.memoized import memoized

logger = logging.getLogger(__name__)


AUTOSETUP_CONF_DIR = u'/etc/tionix/autosetup/'
INSTALL_HISTORY_FILE = os.path.join(AUTOSETUP_CONF_DIR, u'install_history.txt')


def get_last_used_install_source():
    source = ''
    try:
        with codecs.open(INSTALL_HISTORY_FILE, 'r', encoding="utf-8") as f:
            last_line = f.readlines()[-1]
            last_line = last_line.rstrip('\n')
            parsed = last_line.split(';')
            if len(parsed) == 2:
                source = parsed[1]
    except Exception as err:
        logger.error(u'Unable to read PYPI list file: %s'
                     u'', unicode(str(err), "utf-8"))
    return source


def get_latest_rpm_pkg_version(rpm_package_name):
    version = None
    if rpm_package_name:
        cmd = 'apt-cache policy %s' % rpm_package_name

        p = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
        stdout, stderr = p.communicate()

        warning_msg = 'W: '
        if not stderr and stdout and not stdout.startswith(warning_msg):
            latest_available_pkg_index = 2
            version_row = stdout.splitlines()[latest_available_pkg_index]
            version = version_row.split(': ')[-1].split('-')[0]
        else:
            reason = stdout or stderr or u"Unknown"
            logger.error(
                u"Failed to obtain repository for {0}. "
                u"Reason: {1}".format(rpm_package_name.decode('utf-8'),
                                      reason.decode('utf-8'))
            )

    return version


def get_latest_rpm_packages(pip_names):
    versions = {}
    for pip_name in pip_names:
        rpm_name = 'python-module-' + pip_name.replace('-', '_')
        versions[pip_name] = get_latest_rpm_pkg_version(rpm_name)
    return versions


def get_latest_pypi_packages(install_source, packages):
    versions = {}

    err_msg = u'Unable to get packages list from PYPI'
    pypi_index = 'http://{0}/packages/'.format(install_source)

    try:
        response = requests.get(pypi_index, timeout=5)
    except requests.ConnectionError as err:
        logger.error(u'{0}: {1}'.format(err_msg, unicode(err)))
    else:
        status_code = response.status_code
        if status_code == 200:
            html_tree = html.fromstring(response.content)
            for link in html_tree.xpath('//a/text()'):
                package, version = link.split('-')
                package = package.replace('_', '-')
                if package in packages:
                    versions[package] = version.replace('.tar.gz', '')
        else:
            logger.error(u'{0}. Host returns status {1}.'
                         u''.format(err_msg, status_code))
    return versions


def get_actual_versions(packages):
    versions = {package: None for package in packages}
    install_source = get_last_used_install_source()

    if install_source.startswith('apt-get'):
        versions.update(get_latest_rpm_packages(packages))
    elif install_source:
        versions.update(get_latest_pypi_packages(install_source, packages))

    return versions


def check_updates(installed_packages, actual=None):
    packages = {package.replace('_', '-'): version
                for package, version in installed_packages.items()}
    actual_versions = actual or get_actual_versions(packages.keys())
    updates = {}
    for package, version in packages.items():
        actual = actual_versions.get(package)
        installed = version
        if installed and actual:
            if LooseVersion(str(actual)) > LooseVersion(str(installed)):
                updates[package] = actual
        else:
            updates[package] = actual
    return updates


def get_latest_package_version(pip_name):
    packages_versions = get_all_latest_packages_versions()
    return packages_versions.get(pip_name)


@memoized
def get_all_latest_packages_versions():
    return get_actual_versions(PACKAGES.pip_names.values())
