# coding: utf-8

"""
Nodes interface.
"""

from urllib import urlencode

# pylint: disable=import-error, relative-import
from six.moves.urllib import parse

from novaclient.base import getid
from novaclient.v2.hypervisors import (
    Hypervisor as NovaHypervisor,
    HypervisorManager as NovaHypervisorManager)

from tionix_client.helpers import TnxFieldMixin


class Hypervisor(TnxFieldMixin, NovaHypervisor):
    HUMAN_ID = True

    def power_on(self):
        self.manager.power_on(self.id)

    def power_off(self):
        self.manager.power_off(self.id)

    def reset(self):
        self.manager.reset(self.id)

    def update(self, node_title=None):
        self.manager.update(self.id, node_title)

    def monitoring_on(self):
        self.manager.monitoring_on(self.tnx_node_name)

    def monitoring_off(self):
        self.manager.monitoring_off(self.tnx_node_name)


# noinspection PyClassHasNoInit
class HypervisorsManager(NovaHypervisorManager):
    resource_class = Hypervisor
    is_alphanum_id_allowed = True
    alias = 'tnx-hypervisors'
    collection_name = 'hypervisors'
    member_name = 'hypervisor'
    member_alias = 'hypervisor'

    def _build_list_request(self, detailed=True, limit=None, offset=None,
                            only_compute_nodes=True, power_device_id=None):
        qparams = {
            'only_compute_nodes': only_compute_nodes
        }

        if limit:
            qparams['limit'] = limit
        if offset:
            qparams['offset'] = offset
        if power_device_id:
            qparams['power_device_id'] = power_device_id

        return '/{0}{1}{2}'.format(
            self.alias,
            '/detail' if detailed else '',
            '?{0}'.format(urlencode(qparams))
        )

    # pylint: disable=arguments-differ
    def list(self, detailed=True, limit=None, offset=None,
             only_compute_nodes=True, power_device_id=None):
        return self._list(
            self._build_list_request(limit=limit, offset=offset,
                                     only_compute_nodes=only_compute_nodes,
                                     power_device_id=power_device_id),
            self.collection_name)

    def update(self, identifier, **kwargs):
        """
        update node
        kwargs:
            - node_title
            - assigned_to
        """
        return self._update(
            '/{0}/{1}'.format(self.alias, identifier),
            {
                self.member_alias: {
                    key: value
                    for key, value in kwargs.iteritems()
                }
            },
            self.member_name)

    def power_on(self, identifier):
        return self._power_manage_action(identifier, 'power_on')

    def power_off(self, identifier):
        return self._power_manage_action(identifier, 'power_off')

    def reset(self, identifier):
        return self._power_manage_action(identifier, 'reset')

    def _power_manage_action(self, identifier, power_action):
        return self._action(
            'power-manage', identifier, {'power-action': power_action})

    def assign_node_image(self, identifier, image_id):
        return self._action(
            'assign-image', identifier, {'image-id': image_id})

    def set_node_reserved_flag(self, identifier, flag):
        return self._action(
            'reserve', identifier, {'is_reserved': flag})

    def _action(self, action, identifier, info=None, **kwargs):
        """
        Perform a host "action".
        """
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = '/{0}/{1}/action'.format(self.alias, identifier)
        return self.api.client.post(url, body=body)

    def acpi_shutdown(self, identifier, hard=False):
        return self._power_manage_action(
            identifier,
            'ACPI POWER OFF' if hard else 'ACPI SHUT DOWN')

    def set_node_reservs(self, node_name, reserv_node_names):
        return self._action(
            'assign-node-reserve',
            node_name,
            {
                'node-names': reserv_node_names
            })

    def get_reserv_nodes(self, node_name):
        return self._action('get-reserve-nodes', node_name)

    def get_power_actions(self, node_name):
        _resp, body = self.api.client.get(
            u"/{0}/{1}/power-actions".format(
                self.alias,
                node_name))
        return body['power_actions']

    def get_power_action(self, node_name, action_id):
        _resp, body = self.api.client.get(
            u"/{0}/{1}/power-actions/{2}".format(
                self.alias,
                node_name,
                action_id))
        return body['power_action']

    def set_default_power_device_for_node(self, node_name, power_device_id):
        return self._action(
            'set-default-power-control-for-node',
            node_name,
            {
                'power_control_id': power_device_id
            }
        )

    def set_power_device_port_for_node(self, node_id, power_device_id,
                                       power_device_port):
        return self._action(
            'set-power-control-port-for-node',
            node_id,
            {
                'power_control_id': power_device_id,
                'power_control_port': power_device_port
            }
        )

    def count(self, only_compute_nodes=True, power_device_id=None):

        qparams = {
            'only_compute_nodes': only_compute_nodes
        }

        if power_device_id:
            qparams['power_control_id'] = power_device_id

        _resp, body = self.api.client.get(
            u'/{0}/count{1}'.format(self.alias,
                                    u'?{0}'.format(urlencode(qparams)))
        )
        return body['count']

    def host_names(self, names):
        """get_by_hypervisors_host_names"""
        if not isinstance(names, (tuple, list, set)):
            names = [names]

        _resp, body = self.api.client.get(
            u'/{0}/host_names{1}'.format(
                self.alias,
                u'?names={0}'.format(
                    u','.join(
                        (name.replace(',', '\,')
                         for name in names if name is not None)
                    ) if names else u'')))
        return body['hypervisors']

    def get(self, hypervisor):
        """
        Get a specific hypervisor.
        """
        return self._get(
            "/{0}/{1}".format(self.alias, getid(hypervisor)), "hypervisor")

    def search(self, hypervisor_match, servers=False):
        """
        Get a list of matching hypervisors.

        :param hypervisor_match: search matching param.
        :param servers: If True, server information is also retrieved.
        """
        url = ('/{0}/{1}/{2}'.format(
            self.alias, parse.quote(hypervisor_match, safe=''),
            'servers' if servers else 'search'))
        return self._list(url, 'hypervisors')

    def reset_all(self, hypervisor_ids):
        """Reset hypervisors"""
        return self.api.client.get(
            '/{0}/reset?{1}'.format(
                self.alias,
                urlencode({'ids': u','.join(map(str, hypervisor_ids))})))

    def assign_power_device(self, node_id, power_device_id, node_port,
                            is_default=False):
        return self._action(
            'assign-power-control',
            node_id,
            {'power_control_id': power_device_id,
             'node_port': node_port,
             'is_default': is_default}
        )

    def unassign_power_device(self, node_id, power_device_id):
        return self._action(
            'unassign-power-control',
            node_id,
            {'power_control_id': power_device_id}
        )

    def create(self, **kwargs):
        """Create new host"""
        return self._create(
            '/{0}'.format(self.alias),
            {
                self.member_alias: {
                    key: value
                    for key, value in kwargs.iteritems()
                }
            },
            self.member_name)

    def delete(self, hypervisor):
        """Delete host"""
        self._delete('/{0}/{1}'.format(self.alias, getid(hypervisor)))


class HypervisorNovaCliExt(object):
    name = 'tnx_hypervisors'
    manager_class = HypervisorsManager
