# -*- coding: utf-8 -*-
from __future__ import absolute_import

# pylint: disable=wrong-import-position, import-error, wrong-import-order
# noinspection PyPep8Naming
from ldap.controls import knownLDAPControls as controls
from oslo_log import log
from oslo_config import cfg
from pyasn1.type import tag, namedtype, namedval, univ, constraint
from pyasn1.codec.ber import decoder


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class PasswordPolicyWarning(univ.Choice):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('timeBeforeExpiration', univ.Integer().subtype(
            implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
        namedtype.NamedType('graceAuthNsRemaining', univ.Integer().subtype(
            implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1))),
    )


class PasswordPolicyError(univ.Enumerated):
    namedValues = namedval.NamedValues(
        ('passwordExpired', 0),
        ('accountLocked', 1),
        ('changeAfterReset', 2),
        ('passwordModNotAllowed', 3),
        ('mustSupplyOldPassword', 4),
        ('insufficientPasswordQuality', 5),
        ('passwordTooShort', 6),
        ('passwordTooYoung', 7),
        ('passwordInHistory', 8)
    )
    subtypeSpec = (univ.Enumerated.subtypeSpec +
                   constraint.SingleValueConstraint(0, 1, 2, 3, 4, 5, 6, 7, 8))


class PasswordPolicyResponseValue(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType(
            'warning', PasswordPolicyWarning().subtype(
                implicitTag=tag.Tag(tag.tagClassContext,
                                    tag.tagFormatSimple, 0)),),
        namedtype.OptionalNamedType(
            'error', PasswordPolicyError().subtype(
                implicitTag=tag.Tag(tag.tagClassContext,
                                    tag.tagFormatSimple, 1))),
    )


# noinspection PyPep8Naming
class PasswordPolicy(object):
    controlType = '1.3.6.1.4.1.42.2.27.8.5.1'
    error = None
    warning = None

    # noinspection PyUnusedLocal
    def __init__(self, controlType=None, criticality=False, controlValue=None,
                 encodedControlValue=None):
        # TODO: Here should be a class property
        self.__class__.error = None
        self.__class__.warning = None
        self.criticality = criticality
        if controlValue or encodedControlValue:
            self.controlValue = controlValue or self.decodeControlValue(
                encodedControlValue)

    # noinspection PyMethodMayBeStatic
    def encodeControlValue(self, value):
        return value

    def getEncodedTuple(self):
        return self.controlType, self.criticality, None

    # noinspection PyMethodMayBeStatic
    def _hasValue(self, value):
        return value is not None and value.hasValue()

    def getErrorCode(self):
        return self.__class__.error

    def getWarning(self):
        return self.__class__.warning

    # noinspection PyAttributeOutsideInit
    def decodeControlValue(self, encodedControlValue):
        ppolicyValue, _ = decoder.decode(
            encodedControlValue, asn1Spec=PasswordPolicyResponseValue())
        warning = ppolicyValue.getComponentByName('warning')
        if warning is not None:
            warnings = ('timeBeforeExpiration', 'graceAuthNsRemaining')
            self.__class__.warning = {}
            for w in warnings:
                value = warning.getComponentByName(w)
                self.__class__.warning[w] = (
                    int(value) if self._hasValue(value) else None)
        error = ppolicyValue.getComponentByName('error')
        if self._hasValue(error):
            error = int(error)
        else:
            error = None
        self.__class__.error = error


controls[PasswordPolicy.controlType] = PasswordPolicy
