# coding=utf-8
# pylint: disable=import-error, wrong-import-order

from __future__ import absolute_import

import ldap

from oslo_log import log
from oslo_config import cfg

from keystone import exception
# noinspection PyProtectedMember
from keystone.common.ldap.core import (
    KeystoneLDAPHandler, safe_iter, py2ldap, utf8_encode, _get_connection)
from keystone.identity.backends.ldap import Identity, UserApi, GroupApi
from keystone.exception import Unauthorized, ValidationError
# noinspection PyProtectedMember
from keystone.i18n import _

from tionix_client.identity.backends.control import PasswordPolicy


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class PasswordExpired(Unauthorized):
    message_format = '{"error": "PASSWORD_EXPIRED"}'


class PasswordExpireWarning(Unauthorized):
    message_format = ('{"error": "PASSWORD_EXPIRE_WARNING",'
                      '"seconds": %(seconds)s}')


class PasswordAttemptsWarning(Unauthorized):
    message_format = ('{"error": "PASSWORD_ATTEMPTS_WARNING",'
                      '"attempts": %(attempts)s}')


class PasswordConstraintViolation(ValidationError):
    pass


def escape_nt_password(raw_pass):
    """
    Returns special password encoding for Active Directory user password.
    :param raw_pass:
    :return:
    """
    unicode_pass = unicode('"{}"'.format(raw_pass), 'iso-8859-1')
    return unicode_pass.encode('utf-16-le')


class TnxKeystoneLDAPHandler(KeystoneLDAPHandler):

    # user password attribute for Active Directory
    NT_PASSWORD_ATTR = 'unicodePwd'
    PASSWORD_ATTRS = ('userPassword', NT_PASSWORD_ATTR, )

    def modify_s(self, dn, modlist):
        dn_utf8 = utf8_encode(dn)
        ldap_modlist_utf8 = []
        logging_modlist = []

        for op, kind, values in modlist:
            encoded_values = None

            if values is not None:
                encoded_values = []
                for x in safe_iter(values):
                    x = py2ldap(x)
                    if kind == self.NT_PASSWORD_ATTR:
                        encoded_values.append(escape_nt_password(x))
                    else:
                        encoded_values.append(utf8_encode(x))
            if kind == self.NT_PASSWORD_ATTR:
                # default keystone backend behavior for AD password change
                # is sending ADD modify request without DELETE operation.
                # REPLACE operation considers as a password reset by admin.
                # More info:
                # https://msdn.microsoft.com/en-us/library/cc223248.aspx
                op = ldap.MOD_REPLACE  # pylint: disable=no-member

            if kind in self.PASSWORD_ATTRS:
                values = ['****']
            logging_modlist.append((op, kind, values))

            ldap_modlist_utf8.append((op, kind, encoded_values))

        LOG.debug('LDAP modify: dn=%s modlist=%s', dn, logging_modlist)
        return self.conn.modify_s(dn_utf8, ldap_modlist_utf8)


class TnxBaseLdapMixin(object):

    def get_connection(self, user=None, password=None, end_user_auth=False):
        use_pool = self.use_pool
        pool_size = self.pool_size
        pool_conn_lifetime = self.pool_conn_lifetime

        if end_user_auth:
            if not self.use_auth_pool:
                use_pool = False
            else:
                pool_size = self.auth_pool_size
                pool_conn_lifetime = self.auth_pool_conn_lifetime

        conn = _get_connection(self.LDAP_URL, use_pool,
                               use_auth_pool=end_user_auth)

        conn = TnxKeystoneLDAPHandler(conn=conn)

        conn.connect(self.LDAP_URL,
                     page_size=self.page_size,
                     alias_dereferencing=self.alias_dereferencing,
                     use_tls=self.use_tls,
                     tls_cacertfile=self.tls_cacertfile,
                     tls_cacertdir=self.tls_cacertdir,
                     tls_req_cert=self.tls_req_cert,
                     chase_referrals=self.chase_referrals,
                     debug_level=self.debug_level,
                     use_pool=use_pool,
                     pool_size=pool_size,
                     pool_retry_max=self.pool_retry_max,
                     pool_retry_delay=self.pool_retry_delay,
                     pool_conn_timeout=self.pool_conn_timeout,
                     pool_conn_lifetime=pool_conn_lifetime)

        if user is None:
            user = self.LDAP_USER

        if password is None:
            password = self.LDAP_PASSWORD

        # not all LDAP servers require authentication, so we don't bind
        # if we don't have any user/pass
        if user and password:
            ppolicy = PasswordPolicy()
            conn.simple_bind_s(user, password, serverctrls=[ppolicy])
            error = ppolicy.getErrorCode()
            if error is not None:
                if error == 2:  # changeAfterReset
                    raise PasswordExpired()
            warning = ppolicy.getWarning()
            if warning and isinstance(warning, dict):
                seconds = warning.get('timeBeforeExpiration')
                if seconds:
                    raise PasswordExpireWarning(seconds=seconds)
                attempts = warning.get('graceAuthNsRemaining')
                if attempts:
                    raise PasswordAttemptsWarning(attempts=attempts)
        return conn


class TnxUserApi(TnxBaseLdapMixin, UserApi):
    pass


class TnxGroupApi(TnxBaseLdapMixin, GroupApi):
    pass


class TnxIdentity(Identity):

    # noinspection PyArgumentList
    def __init__(self, conf=None):
        super(Identity, self).__init__()  # pylint: disable=bad-super-call
        if conf is None:
            self.conf = CONF
        else:
            self.conf = conf

        self.user = TnxUserApi(self.conf)
        self.group = TnxGroupApi(self.conf)

    def authenticate(self, user_id, password):
        try:
            user_ref = self._get_user(user_id)
        except exception.UserNotFound:
            raise AssertionError(_('Invalid user / password'))
        if not user_id or not password:
            raise AssertionError(_('Invalid user / password'))
        conn = None
        try:
            conn = self.user.get_connection(user_ref['dn'],
                                            password, end_user_auth=True)
            if not conn:
                raise AssertionError(_('Invalid user / password'))
        except ldap.INVALID_CREDENTIALS as err:  # pylint: disable=no-member
            if 'NT_STATUS_PASSWORD_MUST_CHANGE' in unicode(err):
                # Raise 401 http error with understandable message.
                raise PasswordExpired()
            else:
                raise AssertionError(_('Invalid user / password'))
        except (PasswordExpired, PasswordExpireWarning,
                PasswordAttemptsWarning) as err:
            raise err
        except Exception:
            raise AssertionError(_('Invalid user / password'))
        finally:
            if conn:
                conn.unbind_s()
        return self.user.filter_attributes(user_ref)

    def update_user(self, user_id, user):
        try:
            return super(TnxIdentity, self).update_user(user_id, user)
        except ldap.CONSTRAINT_VIOLATION as err:  # pylint: disable=no-member
            if isinstance(err.message, dict):
                msg = err.message.get('info', 'Constraint violation')
            else:
                msg = unicode(err)
            raise PasswordConstraintViolation(msg)
