# coding: utf-8

"""
Build open standard audit information based on incoming requests

AuditMiddleware filter should be placed after keystonemiddleware.auth_token
in the pipeline so that it can utilise the information the Identity server
provides.
"""

import logging
import ast
import urlparse

import os
import sys

# noinspection PyPackageRequirements
import oslo_context.context
# noinspection PyPackageRequirements
import webob.dec
# noinspection PyPackageRequirements
from pycadf import (
    reason, eventfactory, cadftype, resource, host,
    cadftaxonomy as taxonomy)

try:
    # noinspection PyUnresolvedReferences
    import oslo_messaging
    # noinspection PyUnresolvedReferences
    from oslo_config import cfg
except ImportError:
    messaging = False
else:
    messaging = True


AUDIT_GET_ACTIONS = (
    'reboot', 'shutdown', 'startup', 'reset', 'restart_hosts')

EVENT_PREFIX = 'tnxaudit'


class AuditMiddleware(object):
    def __init__(self, app, **conf):
        self._application = app
        self._conf = conf
        self._service_type = None
        self._unsafe_keys = ('password', )

        if messaging:
            transport = oslo_messaging.get_transport(cfg.CONF)
            self._notifier = oslo_messaging.Notifier(
                transport, os.path.basename(sys.argv[0]))

    def _emit_audit(self, context, event_type, payload):
        """Emit audit notification

        if oslo.messaging enabled, send notification. if not, log event.
        """

        if messaging:
            self._notifier.info(context, event_type, payload)
        else:
            logging.info(
                'Event type: %(event_type)s, Context: %(context)s, '
                'Payload: %(payload)s', {
                    'context': context,
                    'event_type': event_type,
                    'payload': payload})

    @staticmethod
    def _clean_path(value, target):
        """Clean path if path has json suffix."""
        object_id = target.id
        if object_id != taxonomy.UNKNOWN:
            value = value.replace('/' + object_id, '')

        value = value[value.rfind('/') + 1:]
        return value[:-5] if value.endswith('.json') else value

    def _get_service_type(self, req):
        if self._service_type:
            return self._service_type

        self._service_type = taxonomy.UNKNOWN

        try:
            catalog = ast.literal_eval(
                req.environ['HTTP_X_SERVICE_CATALOG'])
        except KeyError:
            pass
        else:
            for endp in catalog:
                endpoint_urls = endp['endpoints'][0]
                admin_urlparse = urlparse.urlparse(
                    endpoint_urls.get('adminURL', ''))
                public_urlparse = urlparse.urlparse(
                    endpoint_urls.get('publicURL', ''))
                req_url = urlparse.urlparse(req.host_url)
                if (req_url.netloc == admin_urlparse.netloc or
                        req_url.netloc == public_urlparse.netloc):
                    self._service_type = endp['type']
                    break
        return self._service_type

    def _build_target(self, req, response=None):
        """Build target resource."""
        route = req.environ.get('routes.route')
        _url_gen, route_kwargs = req.environ['wsgiorg.routing_args']
        object_id = route_kwargs.get('id') or taxonomy.UNKNOWN
        if response and object_id == taxonomy.UNKNOWN:
            object_id = self._get_target_id_from_response(response)

        collection_name = getattr(route, 'collection_name', taxonomy.UNKNOWN)
        parent_collection_name = (
            getattr(route, 'parent_resource', None) or {}).get(
                'collection_name')
        service_type = self._get_service_type(req)

        if parent_collection_name:
            type_uri = u"service/{}/{}/{}".format(
                service_type, parent_collection_name, collection_name)
        else:
            type_uri = u"service/{}/{}".format(
                service_type, collection_name)

        target = resource.Resource(
            typeURI=type_uri,
            id=object_id)

        keys = set(getattr(route, 'minkeys', route_kwargs)).difference(['id'])
        for k in keys:
            setattr(target, k, route_kwargs.get(k))

        target.collectionName = collection_name
        if parent_collection_name:
            target.parentCollectionName = parent_collection_name
        return target

    @staticmethod
    def _build_initiator(req, context):
        """Build initiator resource."""
        initiator = resource.Resource(
            typeURI=taxonomy.ACCOUNT_USER,
            host=host.Host(address=req.client_addr, agent=req.user_agent),
        )

        user_id = req.environ.get('HTTP_X_USER_ID')
        if user_id:
            initiator.id = user_id

        user_name = req.environ.get('HTTP_X_USER_NAME')
        if user_name:
            initiator.name = user_name

        request_id = context.get('request_id')
        if request_id:
            initiator.request_id = request_id

        parent_request_id = req.headers.get('X-Parent-Request-ID')
        if parent_request_id:
            initiator.parent_request_id = parent_request_id

        project_id = req.environ.get('HTTP_X_PROJECT_ID')
        if project_id:
            initiator.project_id = project_id

        return initiator

    def _get_action(self, req, target):
        """Take a given Request, parse url path to calculate action type.

        Depending on req.method:

        if POST:

        - path ends with 'action', read the body and use as action;
        - path ending doesn't match collection name, assume use update action;
        - request ends with known path, assume is create action;
        - request ends with unknown path, assume is update action.

        if GET:

        - request ends with known path, assume is update action;
        - request ends with unknown path, assume is read action.

        if PUT, assume update action.
        if DELETE, assume delete action.
        if HEAD, assume read action.

        """
        path = req.path[:-1] if req.path.endswith('/') else req.path
        url_ending = self._clean_path(path, target)
        method = req.method

        body = self._get_json_body(req)

        parent_collection_name = getattr(target, 'parentCollectionName', '')
        collection_name = target.collectionName
        if parent_collection_name:
            collection = parent_collection_name + '/' + collection_name
        else:
            collection = collection_name

        parsed_url = urlparse.urlparse(req.path_qs)
        action_params = urlparse.parse_qs(parsed_url.query)

        method_maps = {
            'PUT': taxonomy.ACTION_UPDATE,
            'PATCH': taxonomy.ACTION_UPDATE,
            'DELETE': taxonomy.ACTION_DELETE,
            'HEAD': taxonomy.ACTION_READ
        }

        if method in method_maps:
            action = '{0}/{1}'.format(method_maps.get(method), collection)
        elif method == 'POST':
            if url_ending == 'action':
                if body:
                    body_action = body.keys()[0]
                    action = '{0}/{1}/{2}'.format(
                        taxonomy.ACTION_UPDATE, collection,
                        body_action)
                else:
                    action = '{0}/{1}'.format(
                        taxonomy.ACTION_CREATE, collection)

            elif url_ending == collection_name:
                action = '{0}/{1}'.format(
                    taxonomy.ACTION_CREATE, collection)
            else:
                action = '{0}/{1}/{2}'.format(
                    taxonomy.ACTION_UPDATE, collection,
                    url_ending)
        elif method == 'GET':
            if url_ending in AUDIT_GET_ACTIONS:
                action = '{0}/{1}/{2}'.format(
                    taxonomy.ACTION_UPDATE, collection,
                    url_ending)
            else:
                action = '{0}/{1}'.format(
                    taxonomy.ACTION_READ, collection)
        else:
            action = '{0}/{1}'.format(
                taxonomy.UNKNOWN, collection)

        if body:
            body_action = body.keys()[0]
            body_params = body[body_action]
            if isinstance(body_params, dict):
                action_params.update(body_params)

        for k in set(action_params).intersection(self._unsafe_keys):
            action_params[k] = '****'

        return action, action_params

    @staticmethod
    def _get_json_body(obj):
        body = {}
        try:
            body = obj.json
        except ValueError:
            pass
        else:
            if not isinstance(body, dict):
                body = {}
        return body

    @staticmethod
    def _get_inner_dict(obj):
        inner = {}
        if obj and isinstance(obj, dict):
            inner = obj[obj.keys()[0]]
            if not isinstance(inner, dict):
                inner = {}
        return inner

    def _get_target_id_from_response(self, response):
        target_id = None
        json_resp = self._get_json_body(response)
        if json_resp:
            target_id = json_resp.get('id') or json_resp.get('tnx_id')
            if not target_id:
                inner = self._get_inner_dict(json_resp)
                target_id = inner.get('id') or inner.get('tnx_id')
        if isinstance(target_id, int):
            target_id = unicode(target_id)
        return target_id or taxonomy.UNKNOWN

    def _get_error_message_from_response(self, response):
        inner_body = self._get_inner_dict(self._get_json_body(response))
        message = inner_body.get('message', '')
        return message

    def _process_response(self, req, response=None, error=None):
        context = oslo_context.context.get_current().to_dict()
        target = self._build_target(req, response)
        # noinspection PyTypeChecker
        action, action_params = self._get_action(req, target)
        if action.startswith(taxonomy.ACTION_READ):
            # skip read and list requests
            return

        initiator = self._build_initiator(req, context)

        outcome = taxonomy.UNKNOWN
        event_reason = taxonomy.UNKNOWN

        msg = None

        if response:
            if 200 <= response.status_int < 400:
                outcome = taxonomy.OUTCOME_SUCCESS
            else:
                msg = self._get_error_message_from_response(response)
                outcome = taxonomy.OUTCOME_FAILURE

            event_reason = reason.Reason(
                reasonType='HTTP', reasonCode=str(response.status_int))

        if error:
            event_reason = reason.Reason(
                reasonType='HTTP', reasonCode='500')
            msg = getattr(error, 'message', '')
            outcome = taxonomy.OUTCOME_FAILURE

        if not msg:
            msg = req.environ.get('tnx_audit_message')

        if msg:
            event_reason.reasonMessage = msg

        event = eventfactory.EventFactory().new_event(
            eventType=cadftype.EVENTTYPE_ACTIVITY,
            outcome=outcome,
            action=action,
            initiator=initiator,
            target=target,
            reason=event_reason)

        event.requestPath = req.path_qs
        event.requestMethod = req.method

        if action_params:
            event.actionArguments = action_params

        event_type = '{0}.{1}'.format(
            EVENT_PREFIX, action.replace('/', '.').replace('-', '_'))
        self._emit_audit(
            oslo_context.context.get_admin_context().to_dict(),
            event_type, event.as_dict())

    @webob.dec.wsgify
    def __call__(self, req):
        try:
            response = req.get_response(self._application)
        except Exception as err:
            self._process_response(req, error=err)
            raise

        try:
            self._process_response(req, response=response)
        except Exception as err:
            logging.exception(err)
        return response


def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def audit_filter(app):
        return AuditMiddleware(app, **conf)
    return audit_filter
