import json
import logging
from collections import namedtuple
from datetime import datetime

from keystoneclient.exceptions import ClientException
from tionix_client.helpers.memoized import lru_cache
from tionix_client.os_clients.keystone_client import get_admin_keystone_client

from tionix_client.settings import CONF


logger = logging.getLogger(__name__)


Target = namedtuple(
    'Target', ['object_id', 'object_type', 'action', 'action_arguments'])

Initiator = namedtuple(
    'Initiator', ['user_id', 'user_name', 'request_id', 'parent_request_id'])

Action = namedtuple(
    'Action',
    [
        'object_id', 'object_type', 'action', 'action_arguments', 'status',
        'response_code', 'message', 'user_id', 'user_name', 'request_id',
        'parent_request_id', 'event_time'
    ])


DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
CACHE_TTL = 60  # seconds
CACHE_SIZE = 100


class BaseEventParser(object):
    RESOURCE_MAPPING = None
    service_name = None

    @staticmethod
    @lru_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def _resolve_user_name(user_id):
        name = None
        try:
            client = get_admin_keystone_client(CONF)
            name = client.users.get(user_id).name
        except ClientException as err:
            logging.error(u'Failed to resolve user name: %s', err)
        return name

    @classmethod
    def _parse_initiator(cls, payload):
        initiator = None
        initiator_obj = payload.get('initiator')
        if initiator_obj:
            user_id = initiator_obj.get('id')
            user_name = initiator_obj.get('name')
            if not user_name:
                user_name = cls._resolve_user_name(user_id)

            initiator = Initiator(
                user_id=user_id,
                user_name=user_name,
                request_id=initiator_obj.get('request_id'),
                parent_request_id=initiator_obj.get('parent_request_id')
            )
        return initiator

    @classmethod
    def _parse_resource_target(cls, payload, resource_type, mapping):
        raise NotImplementedError

    @classmethod
    def _parse_targets(cls, payload):
        targets = []

        for resource_type, mapping in cls.RESOURCE_MAPPING.items():
            targets = cls._parse_resource_target(
                payload, resource_type, mapping)
            if targets:
                break
        return targets

    @classmethod
    def parse_event(cls, event):
        actions = []
        payload = event.get('payload')
        if not payload:
            return actions

        targets = cls._parse_targets(payload)
        if not targets:
            return actions

        initiator = cls._parse_initiator(payload)
        if not initiator:
            return actions

        reason = payload.get('reason') or {}

        event_time = cls._parse_date(payload.get('eventTime'))

        actions = [
            Action(
                object_id=target.object_id,
                object_type=target.object_type,
                action=target.action,
                action_arguments=(
                    unicode(json.dumps(target.action_arguments))
                    if target.action_arguments else None),
                status=payload.get('outcome'),
                response_code=reason.get('reasonCode'),
                message=reason.get('reasonMessage'),
                user_id=initiator.user_id,
                user_name=initiator.user_name,
                request_id=initiator.request_id,
                parent_request_id=initiator.parent_request_id,
                event_time=event_time
            ) for target in targets]

        return actions

    @staticmethod
    def _format_str(val):
        if isinstance(val, (str, unicode)):
            val = val.lower().replace('-', '_')
        return val

    @staticmethod
    def _parse_date(val):
        date = None
        if val:
            try:
                date = datetime.strptime(val.split('+')[0], DATE_FORMAT)
            except ValueError:
                logger.error(u'Failed to parse date: %s', val)
        return date
