import logging

from novaclient.exceptions import ClientException

from tionix_client.os_clients.nova_client import get_admin_nova_client
from tionix_client.helpers.memoized import lru_cache

from tionix_client.settings import CONF
from tionix_client.journal.parsers.base import (
    BaseEventParser, Target, CACHE_TTL, CACHE_SIZE)


logger = logging.getLogger(__name__)


class CustomActions(object):
    ADD_SCSI_META = 'nova.servers.add_scsi_meta'


@lru_cache(maxsize=1, ttl=CACHE_TTL)
def cached_nova_client():
    client = get_admin_nova_client(CONF)
    return client


@lru_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
def resolve_node_id(node_name):
    client = cached_nova_client()
    node_id = None
    nodes = []
    try:
        nodes = client.tnx_hypervisors.list(only_compute_nodes=False)
    except ClientException as err:
        logger.error(u'Failed to get tnx hypervisors: %s', err)

    for node in nodes:
        if node.hypervisor_hostname == node_name:
            node_id = node.id
            break
    return node_id


def process_create_servers_multiple(targets):
    updated_targets = []
    client = cached_nova_client()
    for target in targets:
        updated_targets.append(target)
        try:
            server = client.servers.get(target.object_id)
            reservation_id = getattr(server, 'OS-EXT-SRV-ATTR:reservation_id')
            servers = client.servers.list(
                search_opts={
                    'reservation_id': reservation_id,
                    'all_tenants': True})
            if servers:
                for s in servers:
                    if s.id != server.id:
                        updated_targets.append(Target(
                            object_id=s.id,
                            object_type=target.object_type,
                            action=target.action,
                            action_arguments=target.action_arguments
                        ))
        except ClientException:
            logger.error(
                u'Failed to process server creation: %s', target.object_id)
    return updated_targets


def process_node_power_manage(targets):
    updated_targets = []
    for target in targets:
        if target.action_arguments:
            power_action = target.action_arguments.get('power-action') or ''
            target = Target(
                object_id=target.object_id,
                object_type=target.object_type,
                action='{0}.{1}'.format(
                    target.action, power_action.lower().replace('-', '_')),
                action_arguments=target.action_arguments
            )
        updated_targets.append(target)
    return updated_targets


def process_node_reserve(targets):
    updated_targets = []
    for target in targets:
        if target.action_arguments:
            reserve = target.action_arguments.get('is_reserved')
            action = 'to_reserve' if reserve else 'from_reserve'
            target = Target(
                object_id=target.object_id,
                object_type=target.object_type,
                action='{0}.{1}'.format(
                    target.action, action),
                action_arguments=target.action_arguments
            )
        updated_targets.append(target)
    return updated_targets


@lru_cache(maxsize=CACHE_SIZE)
def has_scsi_info(metadata, image=False):
    """Check if metadata has scsi info"""
    scsi_meta_properties = {
        'hw_disk_bus': 'scsi',
        'hw_scsi_model': 'virtio-scsi',
    }
    prefix = 'image_' if image else ''
    return all(
        metadata.get(prefix + k) == v
        for k, v in scsi_meta_properties.items()
    )


def process_create_server_metadata(targets):
    updated_targets = []
    for target in targets:
        if has_scsi_info(target.action_arguments, image=True):
            target = Target(
                object_id=target.object_id,
                object_type=target.object_type,
                action=CustomActions.ADD_SCSI_META,
                action_arguments=target.action_arguments)
        updated_targets.append(target)
    return updated_targets


def process_create_servers_scsi_image(targets):
    updated_targets = []
    client = cached_nova_client()
    for target in targets:
        updated_targets.append(target)
        image_id = target.action_arguments.get('imageRef')
        if image_id:
            try:
                image = client.images.get(image_id)
                if has_scsi_info(image.metadata):
                    updated_targets.append(Target(
                        object_id=target.object_id,
                        object_type=target.object_type,
                        action=CustomActions.ADD_SCSI_META,
                        action_arguments=target.action_arguments))
            except ClientException:
                logger.error(
                    u'Falied to process server creation: %s', target.object_id)
    return updated_targets


def process_create_servers_scsi_meta(targets):
    updated_targets = []
    for target in targets:
        updated_targets.append(target)
        metadata = target.action_arguments.get('metadata')
        if metadata and has_scsi_info(metadata):
            updated_targets.append(Target(
                object_id=target.object_id,
                object_type=target.object_type,
                action=CustomActions.ADD_SCSI_META,
                action_arguments=target.action_arguments))
    return updated_targets


class NovaEventParser(BaseEventParser):

    RESOURCE_MAPPING = {
        'server': {
            'collections': ['tnx_servers', 'servers'],
            'aliases': ['server_id', 'instance_id', 'server', 'instance'],
            'postprocessors': {
                'nova.create.servers': [
                    process_create_servers_multiple,
                    process_create_servers_scsi_image,
                    process_create_servers_scsi_meta,
                ],
                'nova.create.servers.metadata': [
                    process_create_server_metadata,
                ],
                'nova.update.servers.rebuild': [
                    process_create_servers_scsi_image,
                ]
            }
        },
        'node': {
            'collections': [
                'tnx_hypervisors', 'os_hypervisors', 'tnx_os_hosts'],
            'aliases': [],
            'custom_aliases': {
                'host': resolve_node_id,
                'hostnames': resolve_node_id,
                'hypervisor_hostname': resolve_node_id
            },
            'postprocessors': {
                'nova.update.tnx_hypervisors.power_manage': [
                    process_node_power_manage,
                ],
                'nova.update.tnx_hypervisors.reserve': [
                    process_node_reserve,
                ],
            }
        },
        'project': {
            'collections': ['tnx_tenants', 'os_quota_sets'],
            'aliases': ['project_id']
        }
    }

    service_name = 'nova'

    @classmethod
    def _parse_resource_target(cls, payload, resource_type, mapping):
        targets = []

        action_name = '{0}.{1}'.format(
            cls.service_name,
            cls._format_str(payload.get('action')).replace('/', '.'))

        resource_collections = mapping.get('collections') or []
        resource_aliases = mapping.get('aliases') or []
        custom_aliases = mapping.get('custom_aliases') or []
        postprocessors = mapping.get('postprocessors', {}).get(action_name)

        unknown = 'unknown'
        target_obj = payload.get('target')
        action_params = payload.get('actionArguments') or {}

        resource_id = None
        if not target_obj:
            return targets
        collection = cls._format_str(target_obj.get('collectionName'))
        if collection in resource_collections:
            obj_id = target_obj.get('id')
            if obj_id and obj_id != unknown:
                resource_id = obj_id

        if not resource_id:
            parent_collection = cls._format_str(
                target_obj.get('parentCollectionName'))
            if parent_collection in resource_collections:
                coll_res_id = target_obj.get('id')
                if coll_res_id and coll_res_id != unknown:
                    action_params.update(
                        {cls._format_str(collection): coll_res_id})
                for alias in resource_aliases:
                    obj_id = target_obj.get(alias)
                    if obj_id and obj_id != unknown:
                        resource_id = obj_id
                        break

        if not resource_id:
            resource_id = cls._parse_aliases(action_params, resource_aliases)

        if not resource_id and custom_aliases:
            resource_id = cls._parse_custom_aliases(
                action_params, custom_aliases)

        if resource_id:
            if isinstance(resource_id, list):
                res_ids = set(resource_id)
            else:
                res_ids = [resource_id]

            targets = [
                Target(
                    object_id=res_id,
                    object_type=resource_type,
                    action=action_name,
                    action_arguments=action_params)
                for res_id in res_ids]

            if postprocessors:
                for postprocess in postprocessors:
                    targets = postprocess(targets)
        return targets

    @classmethod
    def _parse_aliases(cls, params, aliases):
        resource_id = None
        for key, val in params.items():
            if cls._format_str(key) in aliases:
                resource_id = val
                break
        return resource_id

    @classmethod
    def _parse_custom_aliases(cls, params, custom_aliases):
        resource_id = None
        for key, val in params.items():
            if cls._format_str(key) in custom_aliases:
                parser = custom_aliases[cls._format_str(key)]
                if isinstance(val, list):
                    resource_id = []
                    for v in val:
                        res_id = parser(v)
                        if res_id is not None:
                            resource_id.append(res_id)
                else:
                    resource_id = parser(val)
                break
        return resource_id
