# coding: utf-8

import json
import logging

from kombu import BrokerConnection

from tionix_client.core.workers import BaseQueueWorker, ServiceConfig
from tionix_client.journal.manager import JournalManager
from tionix_client.journal.parsers import KeystoneEventParser, NovaEventParser
from tionix_client.settings import CONF, get_rabbit_broker_url

logger = logging.getLogger(__name__)


class Service(object):
    KEYSTONE = 'keystone'
    NOVA = 'nova'

    items = KEYSTONE, NOVA


SERVICE_SETTINGS = {
    Service.KEYSTONE: ServiceConfig(
        vhost=CONF.KEYSTONE_RABBIT_VHOST,
        exchange='keystone',
        topic='notifications',
        event_prefix='identity',
        skip_events=['identity.authenticate'],
        parser=KeystoneEventParser),
    Service.NOVA: ServiceConfig(
        vhost=CONF.NOVA_RABBIT_VHOST,
        exchange='nova',
        topic='notifications',
        event_prefix='tnxaudit',
        skip_events=[],
        parser=NovaEventParser)
}


class AuditWorker(BaseQueueWorker):
    QUEUE_PREFIX = 'tnx_audit'

    def _on_message(self, body, message):
        msg = body.get('oslo.message', None)
        if msg:
            oslo_message = json.loads(msg)
            event_type = oslo_message.get('event_type')
            if (event_type.startswith(self.conf.event_prefix) and
                    event_type not in self.conf.skip_events):
                logger.debug('Received message type: %s', event_type)
                actions = self.conf.parser.parse_event(oslo_message)
                for action in actions:
                    self.log_action(action)

    @staticmethod
    def log_action(action):
        try:
            JournalManager.log_action(
                object_id=action.object_id,
                object_type=action.object_type,
                action=action.action,
                action_arguments=action.action_arguments,
                status=action.status,
                response_code=action.response_code,
                message=action.message,
                user_id=action.user_id,
                user_name=action.user_name,
                request_id=action.request_id,
                parent_request_id=action.parent_request_id,
                event_time=action.event_time,
                is_notification=True)
        except Exception as err:
            logger.error(u'Failed to save journal record: %s', err)
        else:
            logger.info(u'Processed action: %s', action)


def start_worker(service):
    logging.info('Started journal worker for the {0} service.'.format(service))
    service_conf = SERVICE_SETTINGS.get(service)
    broker_url = get_rabbit_broker_url(CONF, vhost=service_conf.vhost)
    with BrokerConnection(broker_url) as conn:
        AuditWorker(conn, service_conf).run()
