# coding: utf-8

import json
import logging

from keystoneauth1.identity import Password, Token
from keystoneauth1.session import Session

from keystoneclient.exceptions import Unauthorized, AuthorizationFailure
from keystoneclient.client import Client as KeystoneClient

try:
    from keystoneclient.v3.client import Client as KeystoneClientV3
except ImportError:
    KeystoneClientV3 = None


from tionix_client.helpers.general import get_from_dict
from tionix_client.settings import CONF


logger = logging.getLogger(__name__)


# pylint: disable=too-many-arguments
def get_keystone_client(token=None,
                        username=None, user_id=None, password=None,
                        domain=None, domain_id=None,
                        tenant=None, tenant_name=None, tenant_id=None,
                        project=None, project_name=None, project_id=None,
                        project_domain_name=None, project_domain_id=None,
                        user_domain_name=None, user_domain_id=None,
                        trust_id=None,
                        auth_url=None,
                        auth_version=None,
                        interface=None):
    """Function to get keystone client by provided params.

    :param token: auth token
    :param user_id: User ID for authentication
    :param username: Username for authentication
    :param password: Password for authentication
    :param domain: Domain name for domain scoping
    :param domain_id: Domain ID for domain scoping
    :param tenant: tenant name
    :param tenant_name: tenant name
    :param tenant_id: tenant id
    :param project_id: project id
    :param project: project name
    :param project_name: project name
    :param project_domain_id: Project's domain ID for project scoping
    :param project_domain_name: Project's domain name for project scoping
    :param user_domain_id: User's domain ID for authentication
    :param user_domain_name: User's domain name for authentication
    :param trust_id: Trust ID for trust scoping
    :param auth_url: service endpoint for authorization
    :param auth_version: plugin version for authorization
    :param interface: interface for identity endpoint resolving
    :return: keystone client
    """
    params = {
        'auth_url': auth_url or CONF.KEYSTONE['auth_url'],
        'version': auth_version or CONF.KEYSTONE['auth_version'],
        'token': token,
        'username': username,
        'user_id': user_id,
        'password': password,
        'trust_id': trust_id,
        'domain_id': domain_id,
        'domain_name': domain,
        'tenant_id': tenant_id or project_id,
        'tenant_name': tenant or tenant_name or project or project_name,
    }

    if username:
        if user_domain_id:
            params['user_domain_id'] = user_domain_id
        else:
            params['user_domain_name'] = (
                user_domain_name or CONF.KEYSTONE['user_domain_name'])

    if params['tenant_name']:
        if project_domain_id:
            params['project_domain_id'] = project_domain_id
        else:
            params['project_domain_name'] = (
                project_domain_name or CONF.KEYSTONE['project_domain_name'])

    if not (token or (username or user_id) and password):
        raise AttributeError(
            u'Need a token or username and password for Keystone authenticate')

    client_params = dict()
    client_params['version'] = params.pop('version')
    if interface:
        client_params['interface'] = interface

    if not token:
        params.pop('token')
        auth_plugin = Password
    else:
        params.pop('username')
        params.pop('user_id')
        params.pop('password')
        params.pop('user_domain_id', None)
        params.pop('user_domain_name', None)
        auth_plugin = Token
    auth = auth_plugin(**params)
    session = Session(auth=auth)
    client_params['session'] = session

    try:
        keystone_cl = KeystoneClient(**client_params)
    except Unauthorized as err:
        try:
            json.loads(err.message)
        except ValueError:
            hide_params = ('password', 'token')
            msg = u'Parameters used for authentication are invalid: \n'
            msg += '\n'.join(
                u'{0}: "{1}"'.format(unicode(k), unicode(
                    v) if k not in hide_params else u'*****')
                for k, v in params.iteritems())
            raise Unauthorized(msg)
        else:
            raise
    else:
        # Old code compatibility
        auth_ref = keystone_cl.session.auth.auth_ref
        keystone_cl.user_id = auth_ref.user_id
        keystone_cl.username = auth_ref.username
        keystone_cl.auth_token = auth_ref.auth_token
        keystone_cl.project_id = auth_ref.project_id

        keystone_cl.auth_user_id = auth_ref.user_id
        keystone_cl.project_name = auth_ref.tenant_name
        keystone_cl.auth_tenant_id = auth_ref.tenant_id
        keystone_cl.auth_domain_id = auth_ref.domain_id

    return keystone_cl


def get_tenants_manager(keystone_cl):
    try:
        manager = keystone_cl.projects
    except AttributeError:
        manager = keystone_cl.tenants
    return manager


def keystone_client(req):
    client = None
    context = req.environ['nova.context']
    token = getattr(context, 'auth_token', None)
    token_info = req.environ.get('keystone.token_info', {})

    domain_id = get_from_dict(token_info, 'token.project.domain.id')
    user_domain_name = get_from_dict(token_info, 'token.user.domain.name')
    project_domain_name = get_from_dict(token_info,
                                        'token.project.domain.name')

    if context.is_admin and domain_id:
        try:
            client = get_keystone_client(
                token=token,
                domain_id=domain_id,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name,
            )
        except (AttributeError, AuthorizationFailure, Unauthorized):
            pass

    if not client:
        tenant_id = getattr(context, 'tenant', None)
        client = get_keystone_client(
            token=token,
            tenant_id=tenant_id,
            user_domain_name=user_domain_name,
            project_domain_name=project_domain_name,)

    return client


# noinspection PyTypeChecker
def is_keystone_v3(keystone_cl):
    return (
        KeystoneClientV3 is not None and
        isinstance(keystone_cl, KeystoneClientV3)
    )


# noinspection PyArgumentList
def get_admin_keystone_client(conf=None, domain=True):
    conf = conf or CONF
    params = {
        'username': conf.KEYSTONE['auth_user'],
        'password': conf.KEYSTONE['auth_password'],
    }
    if domain:
        params.update(domain=conf.KEYSTONE['project_domain_name'])
    else:
        params.update(tenant=conf.KEYSTONE['auth_tenant'])
    return get_keystone_client(**params)
