# coding: utf-8

import logging
import pkg_resources

from novaclient.client import Client as NovaClient

from tionix_client.settings import CONF
from tionix_client.os_clients.base import ClientProxy, KeystoneSessionProxy
from tionix_client.os_clients.keystone_client import get_keystone_client

logger = logging.getLogger(__name__)

nova_client_extensions = [
    ep.resolve() for ep in
    pkg_resources.iter_entry_points('novaclient.extension')]


# pylint: disable=too-many-arguments
def get_nova_client(token=None, username=None, password=None,
                    tenant=None, tenant_id=None, project_domain_name=None,
                    user_domain_name=None, trust_id=None,
                    auth_url=None, auth_version=None,
                    timeout=None,
                    keystone_client=None, parent_request_id=None,
                    user_locale=None):
    """Function to get nova client by ks_client.

    :param token: auth token
    :param username: auth name
    :param password: auth password
    :param tenant: tenant name
    :param tenant_id: tenant id
    :param trust_id: Trust ID for trust scoping
    :param project_domain_name: project_domain_name
    :param user_domain_name: user_domain_name
    :param auth_url: auth_url
    :param auth_version: auth_version
    :param timeout: timeout
    :param keystone_client: keystone_client
    :param parent_request_id: parent request id for all requests through
    :param user_locale: user locale
    current client
    """

    if not keystone_client:
        keystone_client = get_keystone_client(
            token=token,
            username=username,
            password=password,
            tenant=tenant,
            tenant_id=tenant_id,
            user_domain_name=user_domain_name,
            project_domain_name=project_domain_name,
            auth_url=auth_url,
            auth_version=auth_version,
            trust_id=trust_id,
        )
    service_type = CONF.KEYSTONE['compute_service_name']

    additional_headers = {
        'Accept-Language': user_locale,
    }

    if parent_request_id:
        additional_headers['X-Parent-Request-ID'] = parent_request_id

    try:
        client = NovaClient(
            CONF.NOVA_VERSION,
            extensions=nova_client_extensions,
            timeout=timeout,
            session=KeystoneSessionProxy(
                keystone_client.session,
                additional_headers=additional_headers
            ),
            service_type=service_type
        )
    except Exception as err:
        logger.error(unicode(err))
        raise err

    return client


class NovaClientProxy(ClientProxy):

    def init_client(self):
        return get_nova_client(token=self.token, tenant_id=self.tenant_id,
                               **self.kwargs)

    def refresh_client(self):
        self.client = get_nova_client(
            username=CONF.KEYSTONE['auth_user'],
            password=CONF.KEYSTONE['auth_password'],
            trust_id=self.trust_id,
            tenant_id=self.tenant_id)


def get_trusted_nova_client(token=None, tenant_id=None, **kwargs):

    return NovaClientProxy(token, tenant_id, **kwargs)


def get_admin_nova_client(conf):
    return get_nova_client(username=conf.KEYSTONE['auth_user'],
                           password=conf.KEYSTONE['auth_password'],
                           tenant=conf.KEYSTONE['auth_tenant'])


def nova_client(req, project_id=None):
    context = req.environ['nova.context']
    token = getattr(context, 'auth_token', None)
    tenant_id = project_id or getattr(context, 'tenant', None)
    return get_nova_client(token=token, tenant_id=tenant_id)
