# coding: utf-8

import collections
import logging
import os
import sys
import traceback

from difflib import SequenceMatcher
from multiprocessing import Process, Queue
from contextlib import contextmanager
from operator import attrgetter
from StringIO import StringIO
import pkg_resources

from cliff import command
try:
    # noinspection PyProtectedMember
    from cliff import _argparse as argparse
except ImportError:
    # noinspection PyUnresolvedReferences
    from cliff import argparse  # cliff < 2.0
from cliff.help import HelpAction

from tionix_client.settings import CONF, BASE_DIR
from tionix_client.functional import Report
from tionix_client.functional.writers import get_writers
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.osc.exceptions import OptionError
from tionix_client.osc.v1 import TESTS_GROUP


logger = logging.getLogger(__name__)

# ratio 0.81 so that node_control and cloud_control do not match
SIMILARITY_RATIO = 0.81

OSC_SILENT_LEVEL = 1

FAILURE_RET_VAL = 1


def get_osc_opts(opts):
    options = {
        'token': opts.token or None,
        'username': opts.username or CONF.KEYSTONE['auth_user'],
        'password': opts.password or CONF.KEYSTONE['auth_password'],
        'tenant': opts.project_name or CONF.KEYSTONE['auth_tenant'],
        'auth_url': opts.auth_url or CONF.KEYSTONE['auth_url'],
        'project_domain_name': (
            opts.project_domain_name or
            CONF.KEYSTONE['project_domain_name']),
        'user_domain_name': (
            opts.user_domain_name or
            CONF.KEYSTONE['user_domain_name']),
    }
    return options


def get_package_name():
    dirs = reversed(os.path.abspath(BASE_DIR).split('/'))
    try:
        package_name = [dir_name.replace('_', '-') for dir_name in dirs
                        if 'tionix_' in dir_name][0]
    except IndexError:
        package_name = None
    return package_name


def verify_name(question, default="yes"):
    """
    Ask a yes/no question via raw_input() and return their answer.
    :param "question" is a string that is presented to the user.
    :param "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).
    :return true/false
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: {0}".format(default))

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def parse_module_names(available_modules, module_names):
    """
    Checks entered module names. If there is a close match, asks a
    [y/n] prompt to check names
    :param available_modules: list of available modules
    :param module_names: list of selected modules
    :return: module_names: validated list of selected modules
    """
    names = []

    for name in module_names:
        name = name.lower()
        if name not in available_modules:
            for available in available_modules:
                ratio = SequenceMatcher(None, name, available).ratio()
                if ratio >= SIMILARITY_RATIO:
                    query = (_(u'You entered: {0}, did you mean: {1}'
                               u'').format(name, available))
                    if verify_name(question=query, default="yes"):
                        names.append(available)
        else:
            names.append(name)

    if names:
        names = list(set(names))
    else:
        raise OptionError(
            _(u'Did not recognize any module. Choose from: {0}').format(
                available_modules))
    return names


def unload_django():
    for module_name, _val in sys.modules.items():
        if 'django' in module_name.split('.'):
            del sys.modules[module_name]


def run_module_ep(module_ep, queue, group, func_kwargs):
    unload_django()  # unloading django from forked process memory if loaded
    results = None
    func_kwargs = func_kwargs or {}
    try:
        func = module_ep.resolve()
    except Exception as err:
        logger.error(
            u"Could not run entry point {0} for module '{1}': {2}".format(
                group, module_ep.name, unicode(err))
        )
    else:
        try:
            with suppress_stdout():
                results = func(**func_kwargs)
        except Exception as err:
            logger.error(
                u"Error while running entry point {0} for module "
                u"'{1}': {2}".format(group, module_ep.name, unicode(err))
            )
    queue.put(results)


class RunningProcess(object):
    def __init__(self, name, process, queue):
        self.name = name
        self.process = process
        self.queue = queue


def get_async_results(module_names, group, func_kwargs=None):
    from tionix_client.functional.helpers import format_module_name
    pool = []
    for ep in get_sorted_modules(group=group, module_names=module_names):
        name = format_module_name(ep.name)
        queue = Queue()
        process = Process(
            target=run_module_ep,
            args=(ep, queue, group, func_kwargs)
        )
        pool.append(RunningProcess(name, process, queue))
        process.start()

    for running_process in pool:
        running_process.process.join()
        result = running_process.queue.get()
        yield (running_process.name, result)


def get_version():
    """ Get installed package version """
    version = None
    package_name = get_package_name()
    try:
        version = pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        logger.error(u'Package not found: {0}'.format(package_name))
    except (pkg_resources.VersionConflict,
            pkg_resources.ContextualVersionConflict) as err:
        logger.warning(u'Version conflict found: {0}'.format(err))
    return version


def get_formatted_version():
    return u"{0} {1}".format(
        get_package_name(), get_version()
    )


@contextmanager
def suppress_stdout():
    new_stdout = StringIO()
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    logging.disable(logging.ERROR)
    try:
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        logging.disable(logging.NOTSET)


def get_sorted_modules(group=TESTS_GROUP, module_names=None):
    if not isinstance(module_names, collections.Iterable):
        module_names = [module_names]
    entry_points = []
    for module_name in module_names:
        entry_points.extend(
            pkg_resources.iter_entry_points(group=group, name=module_name)
        )
    return sorted(entry_points, key=attrgetter('name'))


def get_available_module_names(group):
    return [entry_point.name for entry_point in get_sorted_modules(group)]


def show_available_modules(group):
    print u'Available modules:'
    print(
        '\n'.join(
            [u'\t{0}'.format(name) for name in get_available_module_names(
                group)]
        )
    )


def set_locale(locale):
    """Change LANG in environment and reload i18 module"""

    default_locale = 'en_US.UTF-8'
    locales_map = {
        'ru': 'ru_RU.UTF-8',
    }
    os_lang = os.environ.get('LANG')
    lang = locales_map.get(locale, default_locale)
    if os_lang != lang:
        os.environ['LANG'] = lang
        i18_module = 'tionix_client.helpers.i18'
        if i18_module in sys.modules:
            reload(sys.modules[i18_module])


def run_modules_command(group, parser, module_names=None,
                        func_kwargs=None, silent=False):
    writers = get_writers(silent=silent)
    module_names = module_names or get_available_module_names(group)

    with Report(writers=writers) as report:
        if module_names:
            report.write(_(u'Processing modules: {0}\n\n'
                           u'').format(u', '.join(module_names)))
            for (module_name, result) in get_async_results(
                    module_names, group, func_kwargs):
                if not result:
                    logger.error(_(u"Couldn't get results for "
                                   u"module: {0}.").format(module_name))
                    continue
                parser.add_result(module_name, result)
            report.write(parser.get_results())
        else:
            report.write(_(u'No modules to process\n'))

    return parser.was_successful


class Command(command.Command):
    def run(self, parsed_args):
        result = FAILURE_RET_VAL
        try:
            result = super(Command, self).run(parsed_args)
        except Exception as err:
            if self.app.options.debug:
                logger.error(traceback.format_exc())
            else:
                logger.error(unicode(err))

        return result

    def take_action(self, parsed_args):
        raise NotImplementedError()

    def get_parser(self, prog_name):
        """Return an :class:`argparse.ArgumentParser`.
        """
        parser = argparse.ArgumentParser(
            description=self.get_description(),
            prog=prog_name,
            allow_abbrev=False,
            add_help=False,
        )
        parser.add_argument(
            '-h', '--help',
            action=HelpAction,
            nargs=0,
            default=self,  # tricky
            help=_(u"Show this help message and exit."),
        )
        return parser


class TnxCommand(command.Command):
    def run(self, parsed_args):
        result = FAILURE_RET_VAL
        try:
            result = super(TnxCommand, self).run(parsed_args)
        except Exception as err:
            if self.app.options.debug:
                logger.exception(u"Failed to execute command: %s", err)
            else:
                logger.error(unicode(err))

        return result

    def take_action(self, parsed_args):
        raise NotImplementedError()

    def get_parser(self, prog_name):
        """Return an :class:`argparse.ArgumentParser`.
        """
        parser = argparse.ArgumentParser(
            description=self.get_description(),
            prog=prog_name,
            allow_abbrev=False,
            add_help=False,
        )
        parser.add_argument(
            '-h', '--help',
            action=HelpAction,
            nargs=0,
            default=self,  # tricky
            help=_(u"Show this help message and exit."),
        )
        return parser


class TnxTableDefaultFormatterMixin(object):

    @property
    def formatter_default(self):
        return 'tnx_table'
