# coding: utf-8

import logging

from functools import partial

from tionix_client.osc import utils
from tionix_client.osc.helpers import OSC_SILENT_LEVEL


DEFAULT_API_VERSION = '1'

# Required by the OSC plugin interface
API_NAME = 'tnx-client-plugin'
API_VERSION_OPTION = 'os_tnx_client_api_version'
API_VERSIONS = {
    '1': 'tionix_client.osc.v1.client.Client',
}


# Required by the OSC plugin interface
def make_client(instance):
    """Returns a client to the ClientManager

    Called to instantiate the requested client version.  instance has
    any available auth info that may be required to prepare the client.

    :param ClientManager instance: The ClientManager that owns the new client
    """
    _api_version = getattr(instance, '_api_version')
    # pylint: disable=no-member
    plugin_client = utils.get_client_class(
        API_NAME, _api_version[API_NAME], API_VERSIONS)

    client = plugin_client()
    return client


# Required by the OSC plugin interface
def build_option_parser(parser):
    """Hook to add global options

    Called from openstackclient.shell.OpenStackShell.__init__()
    after the builtin parser has been initialized.  This is
    where a plugin can add global options such as an API version setting.

    :param argparse.ArgumentParser parser: The parser object that has been
        initialized by OpenStackShell.
    """
    utf8 = partial(unicode, encoding='utf-8')
    parser.add_argument(
        '--os-tnx-client-api-version',
        metavar='<tnx-client-api-version>',
        type=utf8,
        # pylint: disable=no-member
        default=utils.env('OS_TNX_CLIENT_API_VERSION') or DEFAULT_API_VERSION,
        help=(
            u'TIONIX.Client OSC Plugin API version, default={0}'
            u' (Env: OS_TNX_CLIENT_API_VERSION)'.format(DEFAULT_API_VERSION)
        )
    )
    parser.add_argument(
        '--os-compute-service-name',
        type=utf8,
        # pylint: disable=no-member
        default=utils.env('OS_COMPUTE_SERVICE_NAME'),
        help=u'Compute service name (Env: OS_COMPUTE_SERVICE_NAME)'
    )

    opts, _args = parser.parse_known_args()
    if opts.verbose_level <= OSC_SILENT_LEVEL and not opts.debug:
        logging.disable(logging.WARN)
    return parser
