# coding: utf-8

import datetime
import json
import logging
import os
import sys

try:
    # noinspection PyProtectedMember
    from cliff import _argparse as argparse
except ImportError:
    # noinspection PyUnresolvedReferences
    from cliff import argparse  # cliff < 2.0

from tionix_client.os_clients.keystone_client import get_keystone_client
from tionix_client.osc.exceptions import OptionError, ConnectionError
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.journal.manager import JournalManager
from tionix_client.osc.helpers import (
    get_available_module_names, OSC_SILENT_LEVEL, run_modules_command,
    TnxTableDefaultFormatterMixin, TnxCommand, get_osc_opts)
from tionix_client.osc.parser import ResultParser
from tionix_client.osc.v1 import DB_PURGE_GROUP
from tionix_client.db.sqlalchemy.migration import get_db_engine
from tionix_client.db.purge import purge_journal
from tionix_client.functional import Report
from tionix_client.functional.writers import get_writers
from tionix_client.journal.constants import ObjectType, ActionStatus

available_module_names = get_available_module_names(DB_PURGE_GROUP)
logger = logging.getLogger(__name__)


class DbPurgeParser(ResultParser):

    table_name = _(u'DB purge table')


class RunDbPurge(TnxTableDefaultFormatterMixin, TnxCommand):
    """
    Run TIONIX module's db purge.
    """

    auth_required = False

    @staticmethod
    def _std(write=False):
        if write:
            std = os.fdopen(os.dup(sys.stdout.fileno()), 'w')
        else:
            std = os.fdopen(os.dup(sys.stdin.fileno()))
        return std

    def take_action(self, parsed_args):
        incorrect = parsed_args.incorrect
        age = parsed_args.age
        journal = parsed_args.journal
        verbose = self.app.options.verbose_level > OSC_SILENT_LEVEL

        if not age and journal:
            raise OptionError(_(u'Invalid parameters. Parameter --journal '
                                u'must be used with parameter --age.'))

        connect_opts = get_osc_opts(self.app.options)
        try:
            keystone_cl = get_keystone_client(**connect_opts)
        except Exception as err:
            raise ConnectionError(u'Could not authenticate in Keystone. '
                                  u'{0}'.format(err))
        func_kwargs = {
            'incorrect': incorrect,
            'age': age,
            'verbose': verbose,
            'journal': journal,
        }

        object_id = user_id = keystone_cl.user_id
        user_name = keystone_cl.username
        domain_id = keystone_cl.auth_domain_id
        object_type = ObjectType.User
        action = 'openstack.tnx.db.purge'
        action_arguments = json.dumps(func_kwargs)
        func_kwargs['user_id'] = user_id
        func_kwargs['user_name'] = user_name
        func_kwargs['domain_id'] = domain_id

        vdi_module = 'tnx_vdi_server'
        status = ActionStatus.SUCCESS
        message = ''

        try:
            if age and journal:
                # journal purge
                engine = get_db_engine()
                connection = engine.connect()
                rows_count = purge_journal(connection=connection, age=age)
                writers = get_writers(silent=False)
                with Report(writers=writers) as report:
                    report.write(_(u'{0} journal rows have been archived.'
                                   u'\n').format(rows_count))
                message = (u'{0} journal rows have been archived.'
                           u''.format(rows_count))
                return

            if age:
                # vdi sessions check
                if vdi_module not in available_module_names:
                    logger.warning(
                        _(u"TIONIX.VDIserver is not installed for using "
                          u"<age> flag. Unable to get sessions list."))
                else:
                    with self._std() as stdin, self._std(write=True) as stdout:
                        func_kwargs['std'] = stdin, stdout
                        run_modules_command(
                            group=DB_PURGE_GROUP,
                            parser=DbPurgeParser(),
                            func_kwargs=func_kwargs,
                            module_names=[vdi_module],
                            silent=True
                        )
                        func_kwargs.pop('age')
                        func_kwargs.pop('std')

            run_modules_command(
                group=DB_PURGE_GROUP,
                parser=DbPurgeParser(),
                func_kwargs=func_kwargs,
                module_names=available_module_names)
            message = ('DB purge operation was run for following modules: {}'
                       ''.format(', '.join(available_module_names)))
        except Exception as err:
            status = ActionStatus.FAILURE
            message = unicode(err)
            raise
        finally:
            JournalManager.log_action(object_id, object_type, action,
                                      is_notification=False,
                                      event_time=datetime.datetime.utcnow(),
                                      user_id=user_id,
                                      user_name=user_name,
                                      action_arguments=action_arguments,
                                      status=status,
                                      message=message)

    @staticmethod
    def check_positive_age(value):
        msg = _(u'Invalid <age> value: must be a positive integer value.')
        msg = msg.encode('utf-8')
        try:
            value = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError(msg)
        else:
            if value <= 0:
                raise argparse.ArgumentTypeError(msg)
        return value

    def get_parser(self, prog_name):
        parser = super(RunDbPurge, self).get_parser(prog_name)

        parser.add_argument(
            '--incorrect',
            help=_(u"Move rows which relevant to deleted instances from "
                   u"production tables to shadow tables."),
            action='store_true')
        parser.add_argument(
            '--age',
            help=_(u"Move rows which relevant to instances from production "
                   u"tables which weren't accessed during last <age> or more "
                   u"days."),
            default=None,
            required=False,
            type=self.check_positive_age)
        parser.add_argument(
            '--journal',
            help=_(u"Archiving journal table. Optional parameter. Applies "
                   u"only with the --age option."),
            required=False,
            action='store_true')
        return parser
