# coding: utf-8
import getpass
import logging
import os

from tionix_client.functional import TestParser, Report
from tionix_client.functional.constants import REPORT_DIR
from tionix_client.functional.helpers import build_variables_table
from tionix_client.functional.settings import get_general_config
from tionix_client.functional.writers import get_writers, VersionedFileWriter
from tionix_client.helpers.system_info import get_system_info
from tionix_client.helpers.table import build_wrapped_table
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.os_clients.nova_client import get_nova_client
from tionix_client.osc.helpers import (
    get_osc_opts, get_available_module_names, suppress_stdout,
    get_async_results, show_available_modules, parse_module_names, set_locale,
    Command, OSC_SILENT_LEVEL)
from tionix_client.osc.v1 import TESTS_GROUP

available_module_names = get_available_module_names(TESTS_GROUP)

logger = logging.getLogger(__name__)


# pylint:disable=too-many-arguments
def run_modules_tests(module_names=None, username=None, silent=True,
                      save_to_file=False, file_dir=REPORT_DIR, socket=None,
                      nova_client=None, locale=None, errors_only=False,
                      debug=False, verbose=False):
    """Run functional tests for modules

    Args:
        module_names (list, str, None): Modules to test. Can be list of names,
            single name or None for all modules.
        username (str): Name of current user.
        silent (bool): If True, test results will be printed to stdout.
        save_to_file (bool): If True, test results will be logged to file.
        file_dir (str): Directory, where report file will be created.
        socket: socket object with `send_chunk` method.
        nova_client: nova client instance.
        locale (str): Language of tests output.
        errors_only (bool): Show only errors in results table.
        debug (bool): Write results with full stacktrace.
        verbose (bool): Write results in verbose mode.

    Returns:
        bool: True if all tests passed, False otherwise.

    """

    if locale:
        set_locale(locale)

    if nova_client:
        auth_ref = nova_client.client.session.auth.auth_ref
        os.environ['OS_TOKEN'] = auth_ref.auth_token
        os.environ['OS_PROJECT_ID'] = auth_ref.project_id

    tests_parser = TestParser()
    writers = get_writers(
        silent=silent, save_to_file=save_to_file, file_dir=file_dir,
        socket=socket, username=username)

    names = module_names or get_available_module_names(TESTS_GROUP)

    with Report(writers=writers) as report:
        if names:
            report.write(
                _(u'Running tests for: {0}\n\n').format(u', '.join(names)))

            func_kwargs = {
                'debug': debug
            }

            with suppress_stdout():
                sys_info = get_system_info(nova_client)
            report.write(build_wrapped_table(rows=sys_info))
            report.write(u'\n')
            if verbose:
                general_conf = get_general_config().items()
                report.write(build_variables_table('TIONIX', general_conf, {}))

            for (module_name, test_result) in get_async_results(
                    names, TESTS_GROUP, func_kwargs):
                if not test_result:
                    logger.error(_(u"Couldn't get test results for "
                                   u"module: {0}.").format(module_name))
                    continue
                tests_parser.process_result(
                    test_result,
                    module_name=module_name,
                    nova_client=nova_client,
                )
                if verbose:
                    report.write(tests_parser.get_module_variables())
                report.write(
                    tests_parser.get_module_summary(errors_only=errors_only))
        else:
            report.write(_(u'No modules to test\n'))

        report.write(tests_parser.get_summary())

    report = None
    for writer in writers:
        if isinstance(writer, VersionedFileWriter):
            report = writer.get_content()
            break

    result = {
        'was_successful': tests_parser.total_result.was_successful,
    }

    if report:
        result['report'] = report

    return result


def get_report(username=None):
    """Get last self-diagnostic report for user"""
    writers = get_writers(username=username, save_to_file=True)
    report = None
    for writer in writers:
        if isinstance(writer, VersionedFileWriter):
            report = writer.get_content()
            break
    return report


class RunDiagnostic(Command):
    """Performs TIONIX modules self-diagnostic."""

    auth_required = False

    def get_description(self):
        return _(u"Performs TIONIX modules self-diagnostic.")

    def take_action(self, parsed_args):
        username = getpass.getuser()
        if parsed_args.list:
            show_available_modules(TESTS_GROUP)
            return os.EX_OK

        if parsed_args.show_last:
            last_report = get_report(username=username)
            if last_report:
                print last_report
            return os.EX_OK

        names = parsed_args.module_names

        if names:
            names = parse_module_names(available_module_names, names)

        verbose = self.app.options.verbose_level > OSC_SILENT_LEVEL
        connect_opts = get_osc_opts(self.app.options)
        try:
            nova_client = get_nova_client(**connect_opts)
        except Exception as e:
            nova_client = None
            logger.warning(
                _(u'Not enough parameters to connect '
                  u'to OpenStack. {0}').format(unicode(e))
            )

        result = run_modules_tests(module_names=names,
                                   username=username,
                                   silent=parsed_args.silent,
                                   save_to_file=parsed_args.save_to_file,
                                   file_dir=parsed_args.file_dir,
                                   nova_client=nova_client,
                                   errors_only=parsed_args.errors_only,
                                   debug=self.app.options.debug,
                                   verbose=verbose)

        was_successful = result.get('was_successful')
        return os.EX_OK if was_successful else os.EX_SOFTWARE

    def get_parser(self, prog_name):
        """Return an :class:`argparse.ArgumentParser`.
        """

        parser = super(RunDiagnostic, self).get_parser(prog_name)

        group = parser.add_mutually_exclusive_group()
        help_msg = (_(u'Names of the modules to test. Choose from: '
                      u'{0}').format(available_module_names))
        group.add_argument(
            '-n', '--names',
            help=help_msg,
            dest='module_names',
            nargs='+',
            default=None)
        group.add_argument(
            '-l', '--list',
            help=_(u'List all modules supporting tests'),
            action='store_true')
        group.add_argument(
            '--show-last',
            dest='show_last',
            help=_(u'Show last self-diagnostic report'),
            action='store_true')
        parser.add_argument(
            '--file-dir',
            dest='file_dir',
            help=_(u'Path to self-diagnostic report dir'),
            default=REPORT_DIR)
        parser.add_argument(
            '--silent',
            dest='silent',
            help=_(u'Suppress test results output'),
            action='store_true')
        parser.add_argument(
            '--save-file',
            dest='save_to_file',
            help=_(u'Write results to file'),
            action='store_true')
        parser.add_argument(
            '--errors-only',
            dest='errors_only',
            help=_(u'Show only errors in results table'),
            action='store_true')

        return parser
