from oslo_log import log as logging

from nova import context, objects
from nova.scheduler.filters import BaseHostFilter, utils


LOG = logging.getLogger(__name__)


class SpecifiedTenantIsolation(BaseHostFilter):
    """Isolate tenants in specific aggregates."""

    # Aggregate data and tenant do not change within a request
    run_filter_once_per_request = True

    RUN_ON_REBUILD = False
    FILTER_NAME = 'filter_tenant_id'

    @property
    def tenants(self):
        admin = context.get_admin_context()
        # pylint: disable=no-member
        # noinspection PyUnresolvedReferences
        aggregates = objects.AggregateList.get_all(admin)
        tenants_list = {aggregate.metadata[self.FILTER_NAME]
                        for aggregate in aggregates
                        if self.FILTER_NAME in aggregate.metadata}
        LOG.debug("Filter by tenants list: %s", tenants_list)
        return tenants_list

    def host_passes(self, host_state, filter_properties):
        """
        It filters hosts by project ID. It passes hosts matching the
        filter_tenant_id specified in the Host Aggregate metadata. Use a comma
        to specify multiple project ids.
        """
        tenant_id = filter_properties.project_id

        metadata = utils.aggregate_metadata_get_by_host(host_state,
                                                        key=self.FILTER_NAME)
        passes = False
        if metadata:
            # check if this host takes current tenant
            configured_tenant_ids = metadata.get(self.FILTER_NAME)

            if configured_tenant_ids:
                if tenant_id in configured_tenant_ids:
                    LOG.debug("Host tenant id %s matched. Host passes.",
                              tenant_id)
                    passes = True
                else:
                    LOG.debug("%s fails, aggregate hasn't tenant id. "
                              "Host doesn't pass.", host_state)
            else:
                LOG.debug("No tenant id's defined on host.")
        else:
            # check if other hosts don't take current tenant
            if tenant_id not in self.tenants:
                LOG.debug("There are no aggregates specified to tenant %s. "
                          "Host passes.", tenant_id)
                passes = True
            else:
                LOG.debug("Dedicated tenant %s couldn't use the host. "
                          "Host doesn't pass.", tenant_id)
        return passes
