# coding: utf-8

"""
Metrics info client.
"""

import six

# pylint: disable=import-error, relative-import
from six.moves.urllib import parse

from novaclient.base import getid
from novaclient.v2.servers import (
    ServerManager as NovaServerManager,
    Server as NovaServer)

from tionix_client.helpers import TnxFieldMixin


TNX_CREATE_ATTRS = ('tnx_type', 'tnx_tenant_id')


class ServerResource(TnxFieldMixin, NovaServer):

    def metrics(self):
        return self.manager.metrics(self)

    def add_user(self, user_id):
        return self.manager.add_user(self.id, user_id)

    def remove_user(self, user_id):
        return self.manager.remove_user(self.id, user_id)

    def add_user_group(self, group_id):
        return self.manager.add_user_group(self.id, group_id)

    def remove_user_group(self, group_id):
        return self.manager.remove_user_group(self.id, group_id)

    def add_backup_templates(self, template_ids):
        return self.manager.add_backup_templates(self.id, template_ids)

    def remove_backup_templates(self, template_ids):
        return self.manager.remove_backup_templates(self.id, template_ids)


# noinspection PyClassHasNoInit
class ServerManager(NovaServerManager):
    resource_class = ServerResource

    server_alias = 'servers'
    tnx_metrics_alias = 'tnx-metrics'

    def list(self, detailed=True, search_opts=None, marker=None, limit=None,
             sort_keys=None, sort_dirs=None):
        if search_opts is None:
            search_opts = {}

        qparams = {}

        for opt, val in six.iteritems(search_opts):
            if val:
                if isinstance(val, six.text_type):
                    val = val.encode('utf-8')
                qparams[opt] = val

        detail = ""
        if detailed:
            detail = "/detail"

        result = []
        while True:
            if marker:
                qparams['marker'] = marker

            if limit and limit != -1:
                qparams['limit'] = limit

            # Transform the dict to a sequence of two-element tuples in fixed
            # order, then the encoded string will be consistent in Python 2&3.
            if qparams or sort_keys or sort_dirs:
                # sort keys and directions are unique since the same parameter
                # key is repeated for each associated value
                # (ie, &sort_key=key1&sort_key=key2&sort_key=key3)
                items = list(qparams.items())
                if sort_keys:
                    # noinspection PyTypeChecker
                    items.extend(('sort_key', sort_key)
                                 for sort_key in sort_keys)
                if sort_dirs:
                    # noinspection PyTypeChecker
                    items.extend(('sort_dir', sort_dir)
                                 for sort_dir in sort_dirs)
                new_qparams = sorted(items, key=lambda x: x[0])
                query_string = "?%s" % parse.urlencode(new_qparams)
            else:
                query_string = ""

            servers = self._list(
                "/{0}{1}{2}".format(
                    self.server_alias, detail, query_string),
                "servers")
            result.extend(servers)

            if not servers or limit != -1:
                break
            marker = result[-1].id
        return result

    def count(self, search_opts=None):
        if search_opts is None:
            search_opts = {}

        search_opts.pop('limit', 0)
        search_opts.pop('marker', 0)
        search_opts.pop('sort_keys', 0)
        search_opts.pop('sort_dirs', 0)

        qparams = {}
        for opt, val in six.iteritems(search_opts):
            if val:
                if isinstance(val, six.text_type):
                    val = val.encode('utf-8')
                qparams[opt] = val

        if qparams:
            items = list(qparams.items())

            new_qparams = sorted(items, key=lambda x: x[0])
            query_string = "?%s" % parse.urlencode(new_qparams)
        else:
            query_string = ""

        servers = self._list(
            "/{0}{1}".format(
                self.server_alias, query_string),
            "servers")

        return len(servers)

    def metrics(self, instance):
        _resp, body = self.api.client.get(
            '/{0}/{1}/{2}'.format(
                self.server_alias, instance.id, self.tnx_metrics_alias))

        return body

    def metrics_names(self):
        _resp, body = self.api.client.get(
            '/{0}'.format(self.tnx_metrics_alias)
        )

        return body

    def recover(self, identifier):
        """ Try to Recover the VM
        :param identifier: VM id
        :return: 202 in success case
        """
        body = {
            'recover': None
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(identifier))
        return self.api.client.post(url, body=body)

    def balance(self, identifier):
        """ Try to balance the VM
        :param identifier: VM id
        :return: 202 in success case
        """
        body = {
            'balance': None
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(identifier))
        return self.api.client.post(url, body=body)

    def repair(self, identifier):
        """ Try to repair the VM
        :param identifier: VM id
        :return: 202 in success case
        """
        body = {
            'repair': None
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(identifier))
        return self.api.client.post(url, body=body)

    def get(self, server):
        """
        Get a server.

        :param server: ID of the :class:`Server` to get.
        :rtype: :class:`Server`
        """
        return self._get(
            "/{0}/{1}".format(self.server_alias, getid(server)),
            "server")

    def add_user(self, server, user_id):
        """ Add user to the VM
        :param server: VM or VM id
        :param user_id: user id
        :return: response with 202 code if success
        """
        body = {
            'add-user': {'user_id': user_id}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def remove_user(self, server, user_id):
        """ Remove user of the VM
        :param server: VM or VM id
        :param user_id: user id
        :return: 202 in success case
        """
        body = {
            'remove-user': {'user_id': user_id}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def reset_with_reboot(self, server, reboot):
        body = {
            'reset-with-reboot': {'reboot': reboot}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def movable(self, server, is_movable):
        body = {
            'movable': {'is_movable': is_movable}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def add_user_group(self, server, group_id):
        """ Add user group to the VM
        :param server: VM or VM id
        :param group_id: group id
        :return: response with 202 code if success
        """
        body = {
            'add-user-group': {'group_id': group_id}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def remove_user_group(self, server, group_id):
        """ Remove user of the VM
        :param server: VM or VM id
        :param group_id: group id
        :return: 202 in success case
        """
        body = {
            'remove-user-group': {'group_id': group_id}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def add_backup_templates(self, server, template_ids):
        """ Add backup templates to the VM
        :param server: VM or VM id
        :param template_ids: backup template ids
        :return: response with 202 code if success
        """
        body = {
            'assign-backup-templates': {'template_ids': template_ids}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)

    def remove_backup_templates(self, server, template_ids):
        """ Remove backup templates of the VM
        :param server: VM or VM id
        :param template_ids: backup template ids
        :return: 202 in success case
        """
        body = {
            'unassign-backup-templates': {'template_ids': template_ids}
        }
        url = '/{0}/{1}/action'.format(self.server_alias, getid(server))
        return self.api.client.post(url, body=body)


class ServerClientExtension(object):
    name = 'tnx_servers'
    manager_class = ServerManager


# This hook adds ability to pass additional VM attributes from novaclient.
# Server creation schema should be extended on API side
def modify_body_for_create(body, **kwargs):
    for tnx_attr_name in TNX_CREATE_ATTRS:
        tnx_attr = kwargs.get(tnx_attr_name)
        if tnx_attr:
            body['server'][tnx_attr_name] = tnx_attr


NovaServerManager.add_hook('modify_body_for_create', modify_body_for_create)
