# coding: utf-8

import copy
import logging

from collections import defaultdict

import yaml
from yaml.constructor import ConstructorError, Constructor, SafeConstructor


logger = logging.getLogger(__name__)

if not logger.handlers:
    # add handler in case when logging hasn't been set up yet
    handler = logging.StreamHandler()
    logger.propagate = 0
    logger.addHandler(handler)

duplicate_keys = set()


def duplicates_detect_constructor(loader, node, deep=False):
    """Detect duplicate keys.
       Doesn't work correctly when the key names in different
       sections the same
    """
    entry = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        value = loader.construct_object(value_node, deep=deep)

        if key in entry:
            duplicate_keys.add(key)

        entry[key] = value
    return loader.construct_mapping(node, deep)


for constructor in (Constructor, SafeConstructor):
    constructor.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                                duplicates_detect_constructor)


# TODO: Refactor to use oslo_config
class Config(object):
    """Configuration settings which may be loaded from config files.

    After loading values are available as attributes of the object.
    """
    _conf = None
    _pass_keys = ('auth_password', 'PASSWORD', 'password')

    is_loaded = False

    def __init__(self, config_files, default_settings, validators,
                 immutable_variables, unused_variables, **kwargs):
        self.default_settings = default_settings
        self.config_files = config_files
        self.validators = validators
        self.immutable_variables = immutable_variables
        self.unused_variables = unused_variables
        self.config_sections = [
            key for key, val in default_settings.items()
            if isinstance(val, dict)]
        self.module_name = kwargs.get('module_name')

    def __getattr__(self, item):
        if not self.is_loaded:
            self.reload()
        return self._conf[item]

    def items(self):
        if not self.is_loaded:
            self.reload()
        return self._conf

    def reload(self, update_loglevel=True):
        self._conf = self._load_configs(
            self.config_files, self.default_settings,
            update_loglevel=update_loglevel)
        self.is_loaded = True

    @staticmethod
    def _read_settings(config_file):
        base_config = {}
        duplicate_keys.clear()
        try:
            with open(config_file) as fh:
                base_config = yaml.safe_load(fh)
        except (EnvironmentError, ConstructorError):
            logger.error(
                u'Failed to read config file: {0}'.format(config_file))

        if duplicate_keys:
            logger.warning(
                u'Duplicate options found in file {0}: {1}. '.format(
                    config_file, list(duplicate_keys)))
        duplicate_keys.clear()

        return base_config

    def _merge_dicts(self, base_dict, update_dict):
        for key, value in update_dict.iteritems():
            base = base_dict.get(key)
            if isinstance(base, dict) and isinstance(value, dict):
                base_dict[key] = self._merge_dicts(base, value)
            else:
                base_dict[key] = value

        return base_dict

    def _load_configs(self, config_files, default_settings,
                      update_loglevel=True):
        result_settings = copy.deepcopy(default_settings)
        loaded_settings = defaultdict(dict)
        temp_settings = defaultdict(dict)
        if isinstance(config_files, (str, unicode)):
            config_files = [config_files]
        for config in config_files:
            temp_settings = self._merge_dicts(
                temp_settings, self._read_settings(config))

        for key, value in temp_settings.iteritems():
            self._update_key(
                loaded_settings, key, value, default_settings,
                log_changes=False)

        # Have to update log level before updating other variables for
        # proper logger output
        log_level = (
            loaded_settings.get('LOG_LEVEL') or
            default_settings.get('LOG_LEVEL'))
        if log_level and update_loglevel:
            logging.getLogger(self.module_name).setLevel(log_level)

        for key, value in loaded_settings.iteritems():
            if key in self.config_sections and value is None:
                continue
            self._update_key(result_settings, key, value, log_changes=True)
        return result_settings

    def _update_key(self, config, key, value, gl=None, parents='',
                    log_changes=False):
        """ Recursively update values of nested dict """
        base_config = gl or config
        if key in base_config:
            if isinstance(value, dict):
                for k, v in value.iteritems():
                    glob = base_config[key] if gl else None
                    formatted_key = '{0}{1}.'.format(parents, key)
                    self._update_key(
                        config[key], k, v, gl=glob,
                        parents=formatted_key, log_changes=log_changes)
            else:
                config[key] = self._get_valid_value(
                    base_config, key, value, parents, log_changes=log_changes)
        elif key not in self.unused_variables:
            logger.error(
                u'{0}{1} is not found in settings.'.format(parents, key))

    def _get_valid_value(self, gl, key, value, parent='', log_changes=False):
        default = gl[key]
        if key in self.immutable_variables:
            value = default
        elif duplicate_keys and key in duplicate_keys:
            logger.error(
                u'Value for duplicate option "{0}{1}" '
                u'set to default: {2} '.format(parent, key, default))
            value = default
        else:
            if key in self.validators:
                value = self.validators[key](key, value, default)
            if default != value and log_changes:
                showed_default = default
                showed_value = value
                if key in self._pass_keys:
                    showed_default = '*' * len(default)
                    showed_value = '*' * len(value)
                logger.debug(u'Changed value of "{0}{1}" from "{2}"'
                             u' to "{3}"'.format(parent, key,
                                                 showed_default,
                                                 showed_value))
        return value
