import logging

from urlparse import urlparse
from keystoneclient import discover


logger = logging.getLogger(__name__)


available_log_levels = ('CRITICAL', 'ERROR', 'WARN', 'WARNING', 'INFO',
                        'DEBUG', 'NOTSET')

db_engines = ('mysql', 'mysql+pymysql')
broker_type = 'amqp'


# noinspection PyUnusedLocal
def get_valid_keystone_auth_url(key, value, default_auth_url):
    parsed_url = urlparse(value)
    invalid_port = False

    try:
        parsed_url.port
    except ValueError:
        invalid_port = True

    if not parsed_url.scheme or not parsed_url.hostname or invalid_port:
        logger.error(
            u'Invalid keystone auth url: {0}. '
            u'Url format for example {1}. '
            u'Keystone auth url set to default: {1}'.format(
                value, default_auth_url))

        value = default_auth_url
    else:
        url = u'{scheme}://{addr}'.format(
            scheme=parsed_url.scheme,
            addr=parsed_url.netloc,
        )
        if url != value:
            logger.debug(
                u"Normalized keystone_auth_url. "
                u"Changed '{0}' to '{1}'".format(value, url)
            )
            value = url
    return value


# noinspection PyUnusedLocal
def get_valid_log_level(key, value, default_log_level):
    processed_value = value.upper()

    if processed_value not in available_log_levels:
        processed_value = default_log_level

        logger.error(
            u'Invalid log level: {0}. '
            u'Log level must be one of {1}. '
            u'Log level set to default: {2}'.format(
                value, available_log_levels, default_log_level))

    return processed_value


# noinspection PyUnusedLocal
def get_valid_cinder_api_version(key, value, default_cinder_version):
    val = unicode(value)
    major_ver = val.replace('v', '').split('.')[0]
    available_major_versions = '1', '2', '3'
    if available_major_versions and major_ver not in available_major_versions:
        logger.error(
            u'Invalid cinder version: {0}. '
            u'Available major versions are {1}. '
            u'Cinder version set to default: {2}'.format(
                value, available_major_versions, default_cinder_version))

        value = default_cinder_version
    return value


# noinspection PyUnusedLocal,PyProtectedMember
def get_valid_keystone_auth_version(key, value, default_auth_version):
    valid = True
    valid_keystone_versions = discover._CLIENT_VERSIONS.keys()
    val = unicode(value)
    try:
        val = float(val.replace('v', '', 1))
    except ValueError:
        valid = False
    else:
        if val not in valid_keystone_versions:
            valid = False

    if not valid:
        logger.error(
            u'Invalid keystone auth version: {0}. '
            u'Keystone auth version set to default: {1}'.format(
                value, default_auth_version))

        value = default_auth_version

    return value


# noinspection PyUnusedLocal
def get_valid_str(key, value, default):
    return unicode(value)


# noinspection PyUnusedLocal
def get_valid_int(key, value, default):
    value = unicode(value)
    if not (value.isdigit() and int(value) > 0):
        logger.error(
            u'Invalid {0}: {1}. '
            u'{0} must consist only of numbers. '
            u'{0} set to default: {2}'.format(key, value, default))

        value = default
    return int(value)


# noinspection PyUnusedLocal
def get_valid_bool(key, value, default):
    if value not in (True, False, 1, 0):
        logger.error(
            u'Invalid {0}: {1}. '
            u'{0} must be boolean type. '
            u'{0} set to default: {2}'.format(key, value, default))

        value = default
    return value


# noinspection PyUnusedLocal
def get_valid_port(key, value=None, default_port=None):
    val = unicode(value)
    if not val.isdigit():
        logger.error(
            u'Invalid port: {0}. '
            u'Port must consist only of numbers. '
            u'Port set to default: {1}'.format(value, default_port))

        value = default_port
    return value


# noinspection PyUnusedLocal
def get_valid_neutron_api_version(key, value, default_neutron_version):

    available_major_versions = ['2']

    val = unicode(value)
    major_version = val.replace('v', '').split('.')[0]
    if available_major_versions and (
            major_version not in available_major_versions):
        logger.error(
            u'Invalid neutron version: {0}. '
            u'Available major versions are {1}. '
            u'Neutron version set to default: {2}'.format(
                value, available_major_versions, default_neutron_version))

        value = default_neutron_version
    return value


# noinspection PyUnusedLocal
def get_valid_db_engine(key, value, default_db_engine):
    if value not in db_engines:
        logger.error(
            u'Invalid DB engine: {0}. '
            u'Supporting DB engines: {1}. '
            u'DB engine set to: {2}'.format(
                value, u", ".join(db_engines), default_db_engine))
        value = default_db_engine
    return value


# noinspection PyUnusedLocal
def get_valid_broker_type(key, value, default):
    if value != broker_type:
        logger.error(
            u'Invalid rabbit broker type: {0}. '
            u'Supporting broker type: "{1}". '
            u'Rabbit broker type set to: {1}'.format(
                value, broker_type))
        value = broker_type
    return value


def get_valid_choice(valid_choices):

    def validator(key, value, default):
        if value not in valid_choices:
            logger.error(
                u'{0} is invalid choice for {1}.'
                u'You should choose one of the following: {2}.'
                u'Value set to default: {3}.'.format(
                    value, key, u', '.join(valid_choices), default
                )
            )
            value = default
        return value

    return validator


validators_map = {
    'LOG_LEVEL': get_valid_log_level,
    'CINDER_VERSION': get_valid_cinder_api_version,
    'NEUTRON_VERSION': get_valid_neutron_api_version,
    'auth_url': get_valid_keystone_auth_url,
    'auth_version': get_valid_keystone_auth_version,
    'auth_password': get_valid_str,
    'password': get_valid_str,
    'PASSWORD': get_valid_str,
    'TRACEBACK_ENABLED': get_valid_bool,
}
