# coding: utf-8
# pylint: disable=attribute-defined-outside-init

import unittest

from keystoneauth1.exceptions.catalog import CatalogException

from tionix_client.os_clients.glance_client import get_glance_client
from tionix_client.os_clients.neutron_client import get_neutron_client
from tionix_client.os_clients.nova_client import get_nova_client
from tionix_client.os_clients.keystone_client import get_keystone_client
from tionix_client.os_clients.cinder_client import get_cinder_client
from tionix_client.helpers.i18 import ugettext as _
from tionix_client.settings import CONF

try:
    from kombu import BrokerConnection
except ImportError:
    BrokerConnection = None


class ServiceType(object):
    CEILOMETER = u'Ceilometer'
    CELERY = u'Celery'
    CINDER = u'Cinder'
    GLANCE = u'Glance'
    DB = u'DB'
    KEYSTONE = u'Keystone'
    NEUTRON = u'Neutron'
    NOVA = u'Nova'
    RABBIT = u'RabbitMQ'


class BaseClientTestCase(unittest.TestCase):

    forbidden_msg = _(u"Test requires to be authorized in OpenStack as admin.")
    endpoint_error_msg = _(u"Couldn't find public endpoint for {service_name} "
                           u"service.")
    extensions_msg = _(u'Extension {service_name} with alias "{alias}" not '
                       u'found.')

    service_connection_msg = _(u'Could not connect to {service_name} with '
                               u'parameters: {params}.')
    objects_list_msg = _(u'Failed to get {objects} list. Connection was to '
                         u'{service_name} with parameters: {params}.')
    attr_not_find_msg = _(u'Attribute {0} not found in configurations.')

    config = CONF

    def _get_auth_params(self):
        try:
            return {
                'username': self.config.KEYSTONE['auth_user'],
                'password': self.config.KEYSTONE['auth_password'],
                'tenant': self.config.KEYSTONE['auth_tenant'],
                'user_domain_name': self.config.KEYSTONE['user_domain_name'],
                'project_domain_name': self.config.KEYSTONE[
                    'project_domain_name']
            }
        except KeyError as err:
            self.fail(self.attr_not_find_msg.format(err))

    def _get_rabbit_params(self):
        try:
            return dict(**self.config.RABBIT_QUEUES)
        except KeyError as err:
            self.fail(self.attr_not_find_msg.format(err))

    def _get_db_params(self):
        try:
            return {
                'engine': self.config.DB['ENGINE'],
                'username': self.config.DB['USER'],
                'password': self.config.DB['PASSWORD'],
                'host': self.config.DB['HOST'],
                'port': self.config.DB['PORT'],
                'name': self.config.DB['NAME']
            }
        except KeyError as err:
            self.fail(self.attr_not_find_msg.format(err))

    rabbit_broker_url = ''

    _cinder_client = None
    _glance_client = None
    _keystone_client = None
    _nova_client = None
    _neutron_client = None
    _rabbit_connection = None

    # noinspection PyBroadException
    @property
    def cinder_client(self):
        if not self._cinder_client:
            try:
                self._cinder_client = get_cinder_client(
                    **self._get_auth_params())
            except CatalogException:
                self.fail_with_params(ServiceType.CINDER,
                                      self.endpoint_error_msg)
            except Exception:
                self.fail_with_params(ServiceType.CINDER,
                                      self.service_connection_msg)
        return self._cinder_client

    # noinspection PyBroadException
    @property
    def glance_client(self):
        if not self._glance_client:
            try:
                self._glance_client = get_glance_client(
                    **self._get_auth_params())
            except Exception:
                self.fail_with_params(ServiceType.GLANCE,
                                      self.service_connection_msg)
        return self._glance_client

    # noinspection PyBroadException
    @property
    def keystone_client(self):
        if not self._keystone_client:
            try:
                self._keystone_client = get_keystone_client(
                    **self._get_auth_params())
            except Exception:
                self.fail_with_params(ServiceType.KEYSTONE,
                                      self.service_connection_msg)
        return self._keystone_client

    # noinspection PyBroadException
    @property
    def neutron_client(self):
        if not self._neutron_client:
            try:
                self._neutron_client = get_neutron_client(
                    **self._get_auth_params())
            except CatalogException:
                self.fail_with_params(ServiceType.NEUTRON,
                                      self.endpoint_error_msg)
            except Exception:
                self.fail_with_params(ServiceType.NEUTRON,
                                      self.service_connection_msg)
        return self._neutron_client

    # noinspection PyBroadException
    @property
    def nova_client(self):
        if not self._nova_client:
            try:
                self._nova_client = get_nova_client(**self._get_auth_params())
            except CatalogException:
                self.fail_with_params(ServiceType.NOVA,
                                      self.endpoint_error_msg)
            except Exception:
                self.fail_with_params(ServiceType.NOVA,
                                      self.service_connection_msg)
        return self._nova_client

    # noinspection PyBroadException
    @property
    def rabbit_connection(self):
        if not self._rabbit_connection:
            if BrokerConnection is None:
                self.fail(u'kombu package is not installed')
            try:
                connection = BrokerConnection(self.rabbit_broker_url)
                self._rabbit_connection = connection.connect()
            except Exception:
                self.fail_with_params(ServiceType.RABBIT,
                                      self.service_connection_msg)
        return self._rabbit_connection

    def check_is_admin_role(self):
        client = self.keystone_client
        data = client.tokens.get_token_data(client.auth_token)
        if not any(role['name'] == 'admin' for role in data['token']['roles']):
            self.skipTest(_(self.forbidden_msg))

    def check_extension(self, alias):
        extensions_url = '/extensions/{0}'
        self.nova_client.client.get(extensions_url.format(alias))

    def fail_with_params(self, service_type, msg, **kwargs):
        self.fail(self.format_msg_with_params(service_type, msg, **kwargs))

    def skip_with_params(self, service_type, msg, **kwargs):
        self.skipTest(self.format_msg_with_params(service_type, msg, **kwargs))

    def format_msg_with_params(self, service_type, msg, **kwargs):
        msg_params = {}
        if service_type == ServiceType.DB:
            params = self._get_db_params().copy()
        elif service_type in (ServiceType.RABBIT, ServiceType.CELERY):
            params = self._get_rabbit_params().copy()
        else:
            params = self._get_auth_params().copy()
        if 'password' in params:
            params['password'] = '******'
        msg_params['service_name'] = service_type
        msg_params['params'] = u', '.join(
            u'{}: "{}"'.format(key, val) for key, val in params.items())
        msg_params.update(kwargs)
        return msg.format(**msg_params)
