# coding: utf-8

import os
import tempfile
from urlparse import urlparse

import requests
from requests.exceptions import RequestException
from requests.status_codes import codes

from keystoneclient.v3.endpoints import VALID_INTERFACES

from tionix_client.helpers.i18 import ugettext as _
from tionix_client.helpers.table import build_wrapped_table
from tionix_client.helpers.updates import (
    INSTALL_HISTORY_FILE, get_actual_versions, get_last_used_install_source,
    get_latest_rpm_pkg_version)
from tionix_client.tests.functional.helpers import (
    get_formatting_distr, check_nova_modules, check_tionix_modules,
    check_tionix_versions)
from tionix_client.tests.functional.base import BaseClientTestCase


class CommonTest(BaseClientTestCase):
    @classmethod
    def setUpClass(cls):

        cls.nova_version = get_formatting_distr('nova')

    def test_tionix_modules_compatibility_with_openstack(self):
        """Check modules compatibility with installed Nova version.

        :lang:ru Проверка совместимости модулей TIONIX с установленной версией
        службы Nova.
        """

        self.assertTrue(self.nova_version, _(u'Nova is not installed.'))

        results = check_tionix_modules(self.nova_version)

        self.assertFalse(results, msg=_(u'Installed version of Nova is not '
                                        u'compatible with one of '
                                        u'TIONIX modules.'))

    def test_tionix_modules_versions(self):
        """Verify that installed TIONIX modules have the same version.

        :lang:ru Проверка совместимости версий установленных модулей TIONIX.
        """

        result = check_tionix_versions()

        self.assertTrue(result, msg=_(u'Installed versions of TIONIX modules '
                                      u'do not match.'))

    def test_valid_openstack_version(self):
        """Verify that all Nova requirements for installed version are
        satisfied.

        :lang:ru Проверка зависимостей службы Nova.
        """

        self.assertTrue(self.nova_version, _(u'Nova is not installed.'))

        msg = None
        results = check_nova_modules(self.nova_version)
        if results:
            file_path = self._save_conflicts(results)
            msg = (
                _(u'Requirements are not satisfied. '
                  u'Please check "{0}".').format(file_path)
            )
        self.assertFalse(results, msg=msg)

    @staticmethod
    def _save_conflicts(results):
        """
        Save requirements conflicts to file

        Returns file path
        """
        file_prefix = u'os_conflicts_'
        headers = (
            u'Module name', u'Version', u'Required version', u'Required by'
        )

        formatted_result = build_wrapped_table(headers, results)

        _handle, file_path = tempfile.mkstemp(prefix=file_prefix)
        with open(file_path, 'w') as fh:
            fh.write(formatted_result.encode('utf8'))
        return file_path

    def test_check_openstack_endpoints(self):
        """Verify that all registered and enabled Openstack services have
        available API endpoints.

        :lang:ru Проверка доступности точек входа API зарегистрированных и
        включенных сервисов Openstack.
        """

        unavailable_endpoints = _(u"Unavailable endpoints: {0}.")
        incomplete_set_endpoints = _(u"Services with an incomplete set of "
                                     u"endpoints: {0}.")
        mandatory_services_msg = _(u"Mandatory service types are not "
                                   u"registered: {0}.")
        mandatory_services = ['compute', 'network', 'image']

        # pylint: disable=no-member
        valid_codes = [codes.MULTIPLE_CHOICES, codes.UNAUTHORIZED, codes.OK]
        services = {}
        service_types = []
        for service in self.keystone_client.services.list():
            if service.enabled:
                services[service.id] = {
                    'type': service.type,
                    'name': service.name, 'endpoints': {}
                }
                service_types.append(service.type)

        mandatory_services_list = []
        for service_type in mandatory_services:
            if service_type not in service_types:
                mandatory_services_list.append(service_type)

        for endpoint in self.keystone_client.endpoints.list():
            if endpoint.enabled:
                services[endpoint.service_id]['endpoints'][
                    endpoint.interface] = endpoint.url

        checked = set()
        unavailable_endpoints_list = []
        incomplete_endpoints_list = []
        for service in services.values():
            service_name = str(service['name']).capitalize()
            for interface in VALID_INTERFACES:
                if interface not in service['endpoints']:
                    incomplete_endpoints_list.append(
                        "{0}:{1}".format(service_name, interface))

            for interface, endpoint in service['endpoints'].items():
                url = urlparse(endpoint)
                endpoint_url = '{0}://{1}'.format(url.scheme, url.netloc)
                if endpoint_url not in checked:
                    checked.add(endpoint_url)
                    try:
                        response = requests.get(endpoint_url)
                    except RequestException:
                        unavailable_endpoints_list.append(
                            "{0}:{1}".format(service_name, interface))
                    else:
                        if response.status_code not in valid_codes:
                            unavailable_endpoints_list.append(
                                "{0}:{1}".format(service_name, interface))

        messages = []
        if mandatory_services_list:
            messages.append(_(mandatory_services_msg).format(
                ', '.join(mandatory_services_list)))
        if unavailable_endpoints_list:
            messages.append(_(unavailable_endpoints).format(
                ', '.join(unavailable_endpoints_list)))
        if incomplete_endpoints_list:
            messages.append(_(incomplete_set_endpoints).format(
                ', '.join(incomplete_endpoints_list)))

        if messages:
            self.fail(' '.join(messages))

    def test_installation_source_file(self):
        """Check whether file with installation sources list exists and
        last source availability.

        :lang:ru Проверка наличия файла со списком источников установки и
            доступности последнего из списка.
        """

        def check_pypi_availability(src):
            pypi_url = 'http://{0}/packages/'.format(src)
            try:
                requests.get(pypi_url)
            except requests.ConnectionError:
                self.fail(_("PyPI {0} is unreachable.").format(pypi_url))

            versions = get_actual_versions(['tionix-client'])
            self.assertIsNotNone(
                versions.get('tionix-client'),
                msg=_("Unable to obtain module version from PyPI."))

        def check_apt_repositories():
            tnx_client_pkg = 'python-module-tionix_client'
            version = get_latest_rpm_pkg_version(tnx_client_pkg)
            if not version:
                self.fail(_('Failed to obtain TIONIX packages from '
                            'apt repositories.'))

        if not os.path.isfile(INSTALL_HISTORY_FILE):
            self.fail(_("File doesn't exist. Make sure that it's "
                        "here: {0}.").format(INSTALL_HISTORY_FILE))

        source = get_last_used_install_source()
        if source.startswith('apt-get'):
            check_apt_repositories()
        else:
            check_pypi_availability(source)
