# coding: utf-8

import os
import pkg_resources

from tionix_client.settings import BASE_DIR


EXCLUDING_PACKAGES = ['tionix-licensing']
TIONIX_KEY = 'tionix'


class Versions(object):
    MITAKA = '13'
    LIBERTY = '12'

    os_requirements = {
        MITAKA: 'requirements_mitaka.txt',
        LIBERTY: 'requirements_liberty.txt'
    }
    tionix_requirements = {
        MITAKA: '>=1.0.0',
        LIBERTY: '<1.0.0',
    }


def get_package_version(distr):
    try:
        package = pkg_resources.get_distribution(distr)
    except (TypeError, pkg_resources.DistributionNotFound):
        return None
    else:
        return package.version


def get_formatting_distr(distr):
    version = get_package_version(distr)
    return version.split('.')[0] if version else ''


def check_tionix_modules(nova_version):
    results = []
    tionix_package_req = Versions.tionix_requirements[nova_version]

    for package in pkg_resources.Environment():
        if TIONIX_KEY in package and package not in EXCLUDING_PACKAGES:
            try:
                pkg_resources.require(package + tionix_package_req)
            except pkg_resources.VersionConflict as ex:
                key, _required_version = get_exception_info(ex)
                if TIONIX_KEY in key:
                    version = get_package_version(package)
                    results.append([key, version, tionix_package_req])

    return results


def check_tionix_versions():
    tionix_versions = []
    result = False

    for package in pkg_resources.Environment():
        if TIONIX_KEY in package and package not in EXCLUDING_PACKAGES:
            try:
                package = pkg_resources.get_distribution(package)
            except TypeError:
                break
            else:
                parsed_version = pkg_resources.parse_version(
                    package.version).base_version.split('.')
                if len(parsed_version) >= 2:
                    version = tuple(parsed_version[:2])
                    if not tionix_versions:
                        tionix_versions.append(version)
                    elif version not in tionix_versions:
                        break
                else:
                    break
    else:
        result = True

    return result


def get_dependencies(nova_version):
    path_to_file = os.path.join(BASE_DIR,
                                Versions.os_requirements[nova_version])
    req_file = None
    try:
        req_file = open(path_to_file, 'rt')
        for dependency in req_file:
            yield dependency.strip().split('#')[0]
    except IOError:
        raise IOError('Error while reading nova requirements file.')
    finally:
        if req_file is not None:
            req_file.close()


def get_exception_info(ex):
    required_version = ''
    key = ex.args[0].key
    if ex.req.specs:
        for specs in ex.req.specs:
            if required_version:
                required_version += ', '
            required_version += specs[0] + specs[1]
    return key, required_version


def get_conflict_dependent_modules(dependency, version):
    dependents = []
    distributions = pkg_resources.working_set
    for distr in distributions:  # pylint: disable=not-an-iterable
        for requirement in distr.requires():
            if requirement.project_name == dependency:
                if not requirement.specifier.contains(version):
                    dependents.append(distr.project_name)
                break
    return dependents


def check_nova_modules(nova_version):
    modules = {}
    for dependency in get_dependencies(nova_version):
        try:
            pkg_resources.require(dependency)
        except (pkg_resources.DistributionNotFound,
                pkg_resources.RequirementParseError):
            pass
        except (pkg_resources.VersionConflict,
                pkg_resources.ContextualVersionConflict) as ex:
            key, required_version = get_exception_info(ex)
            version = get_package_version(key)
            dependents = get_conflict_dependent_modules(key, version)
            if 'nova' not in dependents:
                dependents.insert(0, 'nova')
            modules[key] = (
                key, version, required_version, ', '.join(dependents))
    return modules.values()
