# coding=utf-8

import os
import tempfile
import unittest

import yaml

from tionix_client.settings.loader import Config


BASIC_CONFIG = {
    'NOVA_VERSION': 2,
    'CINDER_VERSION': 2,
    'KEYSTONE': {
        'auth_url': 'http://valid_host:5000',
        'auth_version': '3',
        'auth_user': 'admin',
        'auth_password': 'admin',
        'auth_tenant': 'admin',
        'compute_service_name': 'compute',
        'image_service_name': 'image',
        'network_service_name': 'network',
        'volume_service_name': 'volumev2',
        'ceilometer_service_name': 'metering',
        'user_domain_name': 'default',
        'project_domain_name': 'default'
    },
    'DB': {
        'ENGINE': 'mysql',
        'USER': 'db_user',
        'PASSWORD': 'db_password',
        'HOST': 'localhost',
        'PORT': '3306',
    },
    'RABBIT_QUEUES': {
        'broker_type': 'amqp',
        'host': 'localhost',
        'port': '5672',
        'username': 'user1',
        'password': 'password1',
    },
    'LOG_LEVEL': 'ERROR',
}

MODULE_CONFIG = {
    'DB': {
        'NAME': 'db_name',
    },
    'RABBIT_QUEUES': {
        'vhost': 'test_vhost',
        'username': 'user2',
        'password': 'password2'
    },
    'LOG_LEVEL': 'DEBUG',
    'LATENCY': 100,
}

NOT_SAFE_CONFIG = {
    u'DB': {
        u'ENGINE': u'mysql',
        u'USER': u'not_default_db_user',
        u'PASSWORD': u'not_default_db_password',
        u'HOST': u'not_default_host',
        u'PORT': u'3306',
    }
}

DEFAULT_CONF = {
    'NOVA_VERSION': 2.25,
    'KEYSTONE': {
        'auth_url': 'http://default_host:5000',
        'auth_version': '3',
        'auth_user': 'admin',
        'auth_password': 'admin',
        'auth_tenant': 'admin',
        'compute_service_name': 'compute',
        'volume_service_name': 'volumev2',
        'user_domain_name': 'default',
        'project_domain_name': 'default'
    },
    'DB': {
        'ENGINE': 'mysql',
        'USER': 'default_db_user',
        'PASSWORD': 'default_db_password',
        'HOST': 'localhost',
        'NAME': 'tionix_vdi_server',
        'PORT': '3306'
    },
    'RABBIT_QUEUES': {
        'username': 'default_user',
        'password': 'default_password',
        'host': 'localhost',
        'broker_type': 'amqp',
        'port': '5672',
        'vhost': 'default_vhost',
    },
    'LOG_LEVEL': 'INFO',
    'LATENCY': 60,
}


class TestConfig(unittest.TestCase):
    base_config = module_config = not_safe_config = None

    def setUp(self):
        _fh, self.base_config = tempfile.mkstemp()
        with open(self.base_config, 'w') as fh:
            yaml.dump(BASIC_CONFIG, fh)

        _fh, self.module_config = tempfile.mkstemp()
        with open(self.module_config, 'w') as fh:
            yaml.dump(MODULE_CONFIG, fh)

        self.conf = Config(
            config_files=[self.base_config, self.module_config],
            default_settings=DEFAULT_CONF,
            validators={},
            immutable_variables=['NOVA_VERSION'],
            unused_variables=['CINDER_VERSION'])

    def tearDown(self):
        os.remove(self.base_config)
        os.remove(self.module_config)
        if getattr(self, 'not_safe_config', None):
            os.remove(self.not_safe_config)

    def test_merge_dicts(self):
        base_dict = {}
        update_dict = {}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), base_dict)

        base_dict = {'key_1': {'key_2': 'value_2'}}
        update_dict = {}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), base_dict)

        base_dict = {}
        update_dict = {'key_1': {'key_2': 'value_2'}}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_1': '1'}
        update_dict = {'key_1': {'key_2': 'value_2'}}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_1': {'key_2': 'value_2'}}
        update_dict = {'key_1': 'value_1'}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), update_dict)

        base_dict = {'key_2': 'value_2'}
        update_dict = {'key_1': {'key_2': 'value_2'}}
        after_merge_dict = {'key_1': {'key_2': 'value_2'}, 'key_2': 'value_2'}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), after_merge_dict)

        base_dict = {'key_1': {'key_1': 'value_1'}}
        update_dict = {'key_1': {'key_2': 'value_2'}}
        after_merge_dict = {'key_1': {'key_1': 'value_1', 'key_2': 'value_2'}}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), after_merge_dict)

        base_dict = {
            'key_1': {'key_2': {'key_3': 'value_3', 'key_4': 'value_4'}}}
        update_dict = {
            'key_1': {'key_2': {'key_3': 'new_value_3'}, 'key_3': 'value_3'}}
        after_merge_dict = {
            'key_1': {'key_2': {'key_3': 'new_value_3', 'key_4': 'value_4'},
                      'key_3': 'value_3'}}

        self.assertEqual(
            self.conf._merge_dicts(base_dict, update_dict), after_merge_dict)

    def test_config_loading(self):
        self.assertIsNone(self.conf._conf)
        self.assertEqual(self.conf.LOG_LEVEL, MODULE_CONFIG['LOG_LEVEL'])
        self.assertTrue(isinstance(self.conf._conf, dict))
        self.assertEqual(self.conf.NOVA_VERSION, DEFAULT_CONF['NOVA_VERSION'])
        self.assertEqual(
            self.conf.RABBIT_QUEUES['username'],
            MODULE_CONFIG['RABBIT_QUEUES']['username'])
        self.assertEqual(
            self.conf.KEYSTONE['auth_user'],
            BASIC_CONFIG['KEYSTONE']['auth_user'])

    def test_not_safe_config(self):
        _fh, self.not_safe_config = tempfile.mkstemp()
        with open(self.not_safe_config, 'w') as fh:
            yaml.dump(NOT_SAFE_CONFIG, fh)

        self.conf = Config(
            config_files=[self.not_safe_config],
            default_settings=DEFAULT_CONF,
            validators={},
            immutable_variables=['NOVA_VERSION'],
            unused_variables=['CINDER_VERSION'])

        self.assertFalse(self.conf._read_settings(self.not_safe_config))
        self.assertDictEqual(self.conf.DB, DEFAULT_CONF['DB'])
