# coding: utf-8

import unittest

from datetime import datetime, timedelta

from tionix_client.functional import helpers
from tionix_client.helpers import TnxFieldMixin


class FakeModulesInfo(object):

    def __init__(self, verified=None, expire_at=None):
        self.verified = verified
        self.expire_at = expire_at


class LicenseCheckingTestCase(unittest.TestCase):
    """Case for checking license"""

    @classmethod
    def setUpClass(cls):
        cls.test_name = u'test license validity'
        cls.date_format = '%d.%m.%Y %H:%M:%S'

    def check_license(self, modules_info, test_result, ethalon_result,
                      nova_client=True):
        modules_with_license_info_dict = helpers.modules_info_getters
        helpers.modules_info_getters = {
            helpers.ModuleNames.tnx_cloud_control: lambda x: modules_info
        }
        helpers.add_license_checking_result(
            test_result,

            helpers.ModuleNames.tnx_cloud_control,
            nova_client=nova_client)
        self.assertEqual(ethalon_result, test_result)
        helpers.modules_info_getters = modules_with_license_info_dict

    def test_license_not_installed(self):
        ethalon_result = {
            'tests': {
                'success': [],
                'failures': [(self.test_name, u'License is not installed.')],
                'skipped': [],
            }
        }
        modules_info = FakeModulesInfo()
        test_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [],
            }
        }
        self.check_license(modules_info, test_result, ethalon_result)

    def test_license_not_valid(self):
        ethalon_result = {
            'tests': {
                'success': [],
                'failures': [(self.test_name, u'License has expired.')],
                'skipped': [],
            }
        }
        modules_info = FakeModulesInfo(verified=False)
        test_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [],
            }
        }
        self.check_license(modules_info, test_result, ethalon_result)

    def test_license_valid(self):
        days_delta = 40
        license_date = datetime.now() + timedelta(days_delta)
        license_date = license_date.strftime(self.date_format)
        ethalon_result = {
            'tests': {
                'success': [
                    (self.test_name, '')
                ],
                'failures': [],
                'skipped': [],
            }
        }
        modules_info = FakeModulesInfo(verified=True, expire_at=license_date)
        test_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [],
            }
        }
        self.check_license(modules_info, test_result, ethalon_result)

    def test_license_expire_warning(self):
        days_delta = 10
        license_date = datetime.now() + timedelta(days_delta)
        license_date = license_date.strftime(self.date_format)
        ethalon_result = {
            'tests': {
                'success': [
                    (self.test_name,
                     u'WARNING! License will expire in %d days' % days_delta)
                ],
                'failures': [],
                'skipped': [],
            }
        }
        modules_info = FakeModulesInfo(verified=True, expire_at=license_date)
        test_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [],
            }
        }
        self.check_license(modules_info, test_result, ethalon_result)

    def test_license_without_novaclient(self):
        days_delta = 40
        license_date = datetime.now() + timedelta(days_delta)
        license_date = license_date.strftime(self.date_format)
        ethalon_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [
                    (self.test_name, u"Couldn't connect to OpenStack")],
            }
        }
        modules_info = FakeModulesInfo(verified=True, expire_at=license_date)
        test_result = {
            'tests': {
                'success': [],
                'failures': [],
                'skipped': [],
            }
        }
        self.check_license(modules_info, test_result, ethalon_result,
                           nova_client=False)


class FakeResource(object):

    _loaded = False

    def is_loaded(self):
        return self._loaded

    def get(self):
        setattr(self, 'tnx_test_field', 'test_field_value')
        setattr(self, 'test_field', 'test_field_value')
        self._loaded = True

    def __getattr__(self, k):
        if k not in self.__dict__:
            if not self.is_loaded():
                self.get()
                return self.__getattr__(k)

            raise AttributeError(k)
        else:
            return self.__dict__[k]


class TnxFieldsTestCase(unittest.TestCase):

    def setUp(self):
        self.test_field_value = 'test_field_value'

    def test_tnx_field_available(self):

        class TnxResource(TnxFieldMixin, FakeResource):
            tnx_test_field = self.test_field_value

        resource = TnxResource()
        resource._loaded = True

        self.assertEqual(resource.tnx_test_field, self.test_field_value)

    def test_tnx_field_unavailable(self):
        class TnxResource(TnxFieldMixin, FakeResource):
            pass

        resource = TnxResource()
        resource._loaded = True

        self.assertIsNone(resource.tnx_test_field)

    def test_field_available(self):
        class TnxResource(TnxFieldMixin, FakeResource):
            test_field = self.test_field_value

        resource = TnxResource()
        resource._loaded = True

        self.assertEqual(resource.test_field, self.test_field_value)

    def test_field_unavailable(self):
        class TnxResource(TnxFieldMixin, FakeResource):
            pass

        resource = TnxResource()
        resource._loaded = True

        def test_raise():
            return resource.test_field

        self.assertRaises(AttributeError, test_raise)

    def test_tnx_field_dynamically(self):
        class TnxResource(TnxFieldMixin, FakeResource):
            pass

        resource = TnxResource()

        self.assertEqual(resource.tnx_test_field, self.test_field_value)

    def test_field_dynamically(self):
        class TnxResource(TnxFieldMixin, FakeResource):
            pass

        resource = TnxResource()

        self.assertEqual(resource.test_field, self.test_field_value)
